# SIAF file - I got an excel file from Colin that I saved as a csv

from astropy.io import ascii
import numpy as np

siaf_file = '/ifs/jwst/wit/witserv/data4/nrc/hilbert/simulated_data/NIRCam_SIAF_2017-03-28.csv'

# Read in SIAF
siaf = ascii.read(siaf_file,header_start=1)

# Find the desired row of the SIAF based on
# aperture name (e.g. "NRCA1_FULL")
match = siaf['AperName'] == ap_name
if np.any(match) == False:
    print("Aperture name {} not found in input CSV file.".
          format(aperture))
    sys.exit()

# Extract the row from the SIAF
siaf_row = siaf[match]

# Turn coefficients into functions
sci2idlx, sci2idly, from_units, to_units = read_siaf_table.get_siaf_transform(siaf_row,ap_name,'science','ideal',5)

# Translate detector x,y to x_ideal,y_ideal
# x and y can be single numbers, or lists of numbers
x = [1024,1032]
y = [1024,1000]
x_ideal = sci2idlx(x,y)
y_ideal = sci2idly(x,y)




#! /usr/bin/env python
"""Test the catalog generation functionality in Mirage

Authors
-------
    - Bryan Hilbert

Use
---
    >>> pytest -s test_catalog_generation.py


"""

import copy
import numpy as np
import os

from astropy.io import ascii
import pytest

from mirage.catalogs import catalog_generator
from mirage.catalogs import create_catalog

TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'test_data/')

# Determine if tests are being run on Travis
ON_TRAVIS = 'travis' in os.path.expanduser('~')

if not ON_TRAVIS:
    orig_mirage_data = os.environ['MIRAGE_DATA']
os.environ['MIRAGE_DATA'] = '/test/'


def test_ptsrc_catalog_creation():
    """Test the generation of a basic point source catalog
    """
    ra = np.zeros(10) + 80.0
    dec = np.zeros(10) - 69.8
    mags = np.zeros(10) + 19.

    ptsrc = catalog_generator.PointSourceCatalog(ra=ra, dec=dec)
    ptsrc.add_magnitude_column(mags, instrument='nircam', filter_name='f090w')
    ptsrc.add_magnitude_column(mags, instrument='nircam', filter_name='f200w')

    assert all(ptsrc.ra == ra)
    assert all(ptsrc.dec == dec)
    assert ptsrc.table.colnames == ['index', 'x_or_RA', 'y_or_Dec', 'nircam_f090w_magnitude',
                                    'nircam_f200w_magnitude']


def test_galaxy_catalog_creation():
    """Test the creation of a basic galaxy object catalog
    """
    ra = np.zeros(10) + 80.0
    dec = np.zeros(10) - 69.8
    mags = np.zeros(10) + 19.
    radius = np.zeros(10) + 0.5
    ellip = np.zeros(10) + 0.45
    posang = np.zeros(10) + 27.
    sersic = np.zeros(10) + 3.3

    output_file = os.path.join(TEST_DATA_DIR, 'catalog_generation/galaxy_test.cat')
    gal = catalog_generator.GalaxyCatalog(ra=ra, dec=dec, ellipticity=ellip, radius=radius,
                                          sersic_index=sersic, position_angle=posang, radius_units='arcsec')
    gal.add_magnitude_column(mags, instrument='nircam', filter_name='f090w')
    gal.save(output_file)
    as_read_in = ascii.read(output_file)

    assert gal.table.colnames == ['index', 'x_or_RA', 'y_or_Dec', 'pos_angle', 'sersic_index',
                                  'ellipticity', 'radius', 'nircam_f090w_magnitude']
    assert all(gal.radius == radius)
    assert all(gal.ellipticity == ellip)
    assert all(gal.position_angle == posang)
    assert all(gal.sersic_index == sersic)
    assert all(gal.table == as_read_in)
    os.remove(output_file)


def test_2mass_catalog_generation():
    """Test the generation of a catalog from a 2MASS query
    """
    two_mass_mirage, two_mass_query_results = create_catalog.get_2MASS_ptsrc_catalog(80.4, -69.8, 120)
    comparison_file = os.path.join(TEST_DATA_DIR, 'catalog_generation/TwoMass_test.cat')
    comparison_table = ascii.read(comparison_file)
    assert all(two_mass_mirage.ra == comparison_table['x_or_RA'].data)


def test_catalog_combination():
    """Test the combination of two existing catalogs into one
    """
    two_mass, query_results = create_catalog.get_2MASS_ptsrc_catalog(80.4, -69.8, 120)
    fake_nrc_mags = np.zeros(len(two_mass.ra)) + 16.5
    two_mass.add_magnitude_column(fake_nrc_mags, instrument='nircam', filter_name='f480m',
                                  magnitude_system='vegamag')
    orig_two_mass = copy.deepcopy(two_mass)

    two_mass2, query_results2 = create_catalog.get_2MASS_ptsrc_catalog(70.4, -69.8, 120)
    fake_nrc_mags = np.zeros(len(two_mass2.ra)) + 12.12
    two_mass2.add_magnitude_column(fake_nrc_mags, instrument='niriss', filter_name='f200w',
                                   magnitude_system='vegamag')
    two_mass.add_catalog(two_mass2)

    two_mass2_length = len(two_mass2.ra)
    orig_two_mass_length = len(orig_two_mass.ra)
    assert all(two_mass.ra[(0-two_mass2_length):] == two_mass2.ra)
    assert all(two_mass.dec[(0-two_mass2_length):] == two_mass2.dec)
    assert all(two_mass.ra[0: orig_two_mass_length] == orig_two_mass.ra)
    assert all(two_mass.dec[0: orig_two_mass_length] == orig_two_mass.dec)

    for key in two_mass.magnitudes.keys():
        assert key in ['2mass_j_m_magnitude', '2mass_h_m_magnitude', '2mass_k_m_magnitude',
                       'nircam_f480m_magnitude', 'niriss_f200w_magnitude']


@pytest.mark.skip(reason="Bug with the Besancon model in astroquery.")
def test_besancon_generation():
    """Test the creation of a catalog from a Besancon query
    """
    b_cat, b_query = create_catalog.besancon(80.4, -69.8, 120, coords='ra_dec', email='hilbert@stsci.edu')
    assert len(b_cat.table) > 0
    assert b_cat.table.colnames == ['index', 'x_or_RA', 'y_or_Dec', 'besancon_v_magnitude',
                                    'besancon_j_magnitude', 'besancon_h_magnitude', 'besancon_k_magnitude',
                                    'besancon_l_magnitude']


@pytest.mark.skip(reason="Bug with the Besancon model in astroquery.")
def test_2mass_plus_besaoncon_convenience_function():
    """Test the convenience function that queries 2MASS and Besancon
    and combines the two catalogs
    """
    twomass, query = create_catalog.get_2MASS_ptsrc_catalog(80.4, -69.8, 120)
    both = create_catalog.twoMASS_plus_background(80.4, -69.8, 120, kmag_limits=(16, 25),
                                                  email='hilbert@stsci.edu')
    assert len(twomass.table) < len(both.table)


@pytest.mark.skip(reason="Bug with the Besancon model in astroquery.")
def test_for_proposal():
    """Test the creation of source catalogs from a proposal"""
    xml = os.path.join(TEST_DATA_DIR, 'NIRCam/targets_with_large_separation.xml')
    pointing = xml.replace('.xml', '.pointing')
    output_directory = os.path.join(TEST_DATA_DIR, 'catalog_generation')
    p_cat, g_cat, p_name, g_name, p_map, g_map = create_catalog.for_proposal(xml, pointing,
                                                                             point_source=True,
                                                                             extragalactic=True,
                                                                             catalog_splitting_threshold=0.12,
                                                                             email='hilbert@stsci.edu',
                                                                             out_dir=output_directory,
                                                                             save_catalogs=True,
                                                                             besancon_seed=2345,
                                                                             galaxy_seed=2346)

    ptsrc_map = {'001': 'ptsrc_for_targets_with_large_separation_observations_001.cat',
                 '002': 'ptsrc_for_targets_with_large_separation_observations_002.cat',
                 '003': 'ptsrc_for_targets_with_large_separation_observations_003.cat'}

    galaxy_map = {'001': 'galaxies_for_targets_with_large_separation_observations_001.cat',
                  '002': 'galaxies_for_targets_with_large_separation_observations_002.cat',
                  '003': 'galaxies_for_targets_with_large_separation_observations_003.cat'}

    ptsrc_name = []
    galaxy_name = []
    for key in ptsrc_map:
        ptsrc_name.append(os.path.join(output_directory, ptsrc_map[key]))
        galaxy_name.append(os.path.join(output_directory, galaxy_map[key]))

    assert ptsrc_map == p_map
    assert galaxy_map == g_map
    assert ptsrc_name == p_name
    assert galaxy_name == g_name

    truth_ptsrc = ['ptsrc_1.cat', 'ptsrc_2.cat', 'ptsrc_3.cat']
    truth_galaxy = ['galaxy_1.cat', 'galaxy_2.cat', 'galaxy_3.cat']

    for pcat_new, pcat in zip(p_cat, truth_ptsrc):
        truth_file = os.path.join(TEST_DATA_DIR, 'catalog_generation/{}'.format(pcat))
        truth = ascii.read(truth_file)

        for col in pcat_new.table.colnames:
            assert all(pcat_new.table[col].data == truth[col].data)

    for gcat_new, gcat in zip(g_cat, truth_galaxy):
        truth_file = os.path.join(TEST_DATA_DIR, 'catalog_generation/{}'.format(gcat))
        truth = ascii.read(truth_file)

        for col in gcat_new.table.colnames:
            print('DATA: {}'.format(col))
            print(gcat_new.table[col].data)
            print('')
            print(truth[col].data)
            if 'magnitude' in col:
                assert np.allclose(gcat_new.table[col].data, truth[col].data, rtol=0, atol=0.5)
            else:
                assert all(gcat_new.table[col].data == truth[col].data)

    # Remove the catalog files just produced
    for del_ptsrc, del_galaxy in zip(ptsrc_name, galaxy_name):
        print(os.path.join(output_directory, del_ptsrc), os.path.join(output_directory, del_galaxy))
        os.remove(os.path.join(output_directory, del_ptsrc))
        os.remove(os.path.join(output_directory, del_galaxy))
        cwd = os.getcwd()
        try:
            os.remove(os.path.join(cwd, 'observation_list.yaml'))
            os.remove(os.path.join(cwd, 'expand_for_detectors.csv'))
            os.remove(os.path.join(cwd, 'Observation_table_for_targets_with_large_separation.csv'))
        except:
            pass


@pytest.mark.skip(reason="Bug with the Besancon model in astroquery.")
def test_get_all_catalogs():
    """Test the wrapper that queries anc combines catalogs from all sources"""
    ra = 80.4
    dec = -69.8
    width = 120.
    ins = 'NIRCAM'
    filters = ['F150W', 'F356W', 'F444W', 'F480M']

    cat, headers = create_catalog.get_all_catalogs(ra, dec, width, kmag_limits=(13, 29),
                                                   email='hilbert@stsci.edu', instrument=ins, filters=filters,
                                                   besancon_seed=1234)
    comparison_file = os.path.join(TEST_DATA_DIR, 'catalog_generation/get_all_catalogs.cat')
    comparison_data = ascii.read(comparison_file)

    # Note that if Besancon/WISE/GAIA/2MASS query results change, this will
    # fail without there being a problem with Mirage.
    for col in cat.table.colnames:
        try:
            assert all(cat.table[col].data == comparison_data[col].data), \
                "Retrieved catalog does not match expected."
        except TypeError:
            assert False, "Retrieved catalog does not match expected."


def test_gaia_query():
    """Test the GAIA query and transformation into a Mirage-format catalog"""
    ra = 80.4
    dec = -69.8
    box_width = 200.
    cat, query, gaia_2mass_cross, gaia_wise_cross = create_catalog.get_gaia_ptsrc_catalog(ra, dec, box_width)
    assert len(cat.table) == 1153
    assert cat.table.colnames == ['index', 'x_or_RA', 'y_or_Dec', 'gaia_phot_g_mean_mag_magnitude',
                                  'gaia_phot_bp_mean_mag_magnitude', 'gaia_phot_rp_mean_mag_magnitude']


def test_random_ra_dec_values():
    """Test the random RA, Dec value generator used when getting Besancon
    sources"""
    num_stars = 10
    ra_min = 1.0
    ra_max = 1.1
    dec_min = 40.
    dec_max = 40.1
    ra1, dec1 = create_catalog.generate_ra_dec(num_stars, ra_min, ra_max, dec_min, dec_max,
                                               seed=37465)
    ra2, dec2 = create_catalog.generate_ra_dec(num_stars, ra_min, ra_max, dec_min, dec_max,
                                               seed=37465)
    assert np.all(ra1 == ra2)
    assert np.all(dec1 == dec2)
    assert np.min(ra1) >= ra_min
    assert np.max(ra1) <= ra_max
    assert np.min(dec1) >= dec_min
    assert np.max(dec1) <= dec_max


def test_cat_from_file():
    """Test function for reading ascii catalogs into catalog objects"""
    data_path = os.path.join(TEST_DATA_DIR, 'catalog_generation/')
    catalogs = {'ptsrc_1.cat': ('point_source', catalog_generator.PointSourceCatalog),
                'galaxy_1.cat': ('galaxy', catalog_generator.GalaxyCatalog),
                'extended_test.cat': ('extended', catalog_generator.ExtendedCatalog),
                'moving_point_source_test.cat': ('moving_point_source', catalog_generator.MovingPointSourceCatalog),
                'moving_sersic_source_test.cat': ('moving_sersic', catalog_generator.MovingSersicCatalog),
                'moving_extended_source_test.cat': ('moving_extended', catalog_generator.MovingExtendedCatalog),
                'nonsidereal_source_test.cat': ('non_sidereal', catalog_generator.NonSiderealCatalog)}
    for cat_name in catalogs:
        cat_path = os.path.join(data_path, cat_name)
        cat_object = catalog_generator.cat_from_file(cat_path, catalogs[cat_name][0])
        assert isinstance(cat_object, catalogs[cat_name][1])

if not ON_TRAVIS:
    os.environ['MIRAGE_DATA'] = orig_mirage_data



"""Test the functions provided by spectra_from_catalog.py

Authors
-------
    - Bryan Hilbert

Use
---
    >>> pytest -s test_spectra_from_catalog.py


"""
import os
import numpy as np
import copy

from astropy.io import ascii
from astropy.table import Table
import astropy.units as u
from synphot import SourceSpectrum, Observation, units
from synphot.spectrum import SpectralElement
from synphot.models import Empirical1D
from synphot.models import BlackBodyNorm1D

from mirage.catalogs import spectra_from_catalog as spec
from mirage.catalogs import hdf5_catalog as hdf5
from mirage.utils.constants import FLAMBDA_CGS_UNITS, FNU_CGS_UNITS, MEAN_GAIN_VALUES
from mirage.utils.utils import get_filter_throughput_file, magnitude_to_countrate


TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'test_data/hdf5_catalogs')


def test_get_filter_info():
    nrc = spec.get_filter_info(['nircam_f444w_magnitude'], 'abmag')
    assert nrc == {'nircam_f444w_magnitude': (6.6928e-22 * FLAMBDA_CGS_UNITS, 4.3637e-31 * FNU_CGS_UNITS, 27.3004,
                                              4.4212 * u.micron)}

    nis = spec.get_filter_info(['niriss_f200w_magnitude'], 'vegamag')
    assert nis == {'niriss_f200w_magnitude': (2.173398e-21 * FLAMBDA_CGS_UNITS, 2.879494e-31 * FNU_CGS_UNITS,
                                              26.04898, 1.9930 * u.micron)}

    fgs = spec.get_filter_info(['fgs_magnitude'], 'stmag')
    assert fgs == {'fgs_magnitude': (1.593395e-22 * FLAMBDA_CGS_UNITS, 3.324736e-32 * FNU_CGS_UNITS, 33.39426,
                                     2.5011 * u.micron)}


def test_hdf5_file_input():
    """Case where an hdf5 file is input. One of these also includes a
    normalized spectrum"""
    catfile = os.path.join(TEST_DATA_DIR, 'point_sources.cat')
    output_hdf5 = os.path.join(TEST_DATA_DIR, 'all_spectra.hdf5')
    sed_file = os.path.join(TEST_DATA_DIR, 'sed_file_with_normalized_dataset.hdf5')
    sed_catalog = spec.make_all_spectra(catfile, input_spectra_file=sed_file,
                                        normalizing_mag_column='nircam_f444w_magnitude',
                                        output_filename=output_hdf5, module='A')

    comparison = hdf5.open(os.path.join(TEST_DATA_DIR, 'output_spec_from_hdf5_input_including_normalized.hdf5'))
    constructed = hdf5.open(sed_catalog)
    for key in comparison:
        assert key in constructed.keys()
        assert all(comparison[key]["wavelengths"].value == constructed[key]["wavelengths"].value)
        assert all(comparison[key]["fluxes"].value == constructed[key]["fluxes"].value)
        assert comparison[key]["wavelengths"].unit == constructed[key]["wavelengths"].unit
        assert comparison[key]["fluxes"].unit == constructed[key]["fluxes"].unit

    cat_base = catfile.split('.')[0]
    outbase = cat_base + '_with_flambda.cat'
    flambda_output_catalog = os.path.join(TEST_DATA_DIR, outbase)
    os.remove(flambda_output_catalog)
    os.remove(sed_catalog)


def test_manual_inputs():
    """Case where spectra are input manually along side ascii catalog"""
    # Test case where spectra are input manually
    catfile = os.path.join(TEST_DATA_DIR, 'point_sources.cat')
    output_hdf5 = os.path.join(TEST_DATA_DIR, 'all_spectra.hdf5')
    hdf5file = os.path.join(TEST_DATA_DIR, 'sed_file_with_normalized_dataset.hdf5')
    sed_dict = hdf5.open(hdf5file)
    sed_catalog = spec.make_all_spectra(catfile, input_spectra=sed_dict,
                                        normalizing_mag_column='nircam_f444w_magnitude',
                                        output_filename=output_hdf5, module='A')
    constructed = hdf5.open(sed_catalog)
    comparison = hdf5.open(os.path.join(TEST_DATA_DIR, 'output_spec_from_manual_input.hdf5'))
    for key in comparison:
        assert key in constructed.keys()
        assert all(comparison[key]["wavelengths"].value == constructed[key]["wavelengths"].value)
        assert all(comparison[key]["fluxes"].value == constructed[key]["fluxes"].value)
        assert comparison[key]["wavelengths"].unit == constructed[key]["wavelengths"].unit
        assert comparison[key]["fluxes"].unit == constructed[key]["fluxes"].unit

    cat_base = catfile.split('.')[0]
    outbase = cat_base + '_with_flambda.cat'
    flambda_output_catalog = os.path.join(TEST_DATA_DIR, outbase)
    os.remove(flambda_output_catalog)
    os.remove(sed_catalog)


def test_manual_plus_file_inputs():
    """Case where spectra are input via hdf5 file as well as manually"""
    catfile = os.path.join(TEST_DATA_DIR, 'point_sources.cat')
    sed_file = os.path.join(TEST_DATA_DIR, 'sed_file_with_normalized_dataset.hdf5')
    output_hdf5 = os.path.join(TEST_DATA_DIR, 'all_spectra.hdf5')
    manual_sed = {}
    manual_sed[7] = {"wavelengths": [0.9, 1.4, 1.9, 3.5, 5.1]*u.micron,
                     "fluxes": [1e-17, 1.1e-17, 1.5e-17, 1.4e-17, 1.1e-17] * FLAMBDA_CGS_UNITS}
    sed_catalog = spec.make_all_spectra(catfile, input_spectra=manual_sed, input_spectra_file=sed_file,
                                        normalizing_mag_column='nircam_f444w_magnitude',
                                        output_filename=output_hdf5, module='A')
    comparison = hdf5.open(os.path.join(TEST_DATA_DIR, 'output_spec_from_file_plus_manual_input.hdf5'))
    constructed = hdf5.open(output_hdf5)
    for key in comparison:
        assert key in constructed.keys()
        assert all(comparison[key]["wavelengths"].value == constructed[key]["wavelengths"].value)
        assert all(comparison[key]["fluxes"].value == constructed[key]["fluxes"].value)
        assert comparison[key]["wavelengths"].unit == constructed[key]["wavelengths"].unit
        assert comparison[key]["fluxes"].unit == constructed[key]["fluxes"].unit

    cat_base = catfile.split('.')[0]
    outbase = cat_base + '_with_flambda.cat'
    flambda_output_catalog = os.path.join(TEST_DATA_DIR, outbase)
    os.remove(flambda_output_catalog)
    os.remove(sed_catalog)


def test_multiple_mag_columns():
    """Case where ascii catalog with multiple magnitude columns is input"""
    catfile = os.path.join(TEST_DATA_DIR, 'point_sources.cat')
    output_hdf5 = os.path.join(TEST_DATA_DIR, 'all_spectra.hdf5')
    sed_catalog = spec.make_all_spectra(catfile, output_filename=output_hdf5)
    constructed = hdf5.open(sed_catalog)
    comparison = hdf5.open(os.path.join(TEST_DATA_DIR, 'output_spec_from_multiple_filter.hdf5'))
    for key in comparison:
        assert key in constructed.keys()
        assert all(comparison[key]["wavelengths"].value == constructed[key]["wavelengths"].value)
        assert all(comparison[key]["fluxes"].value == constructed[key]["fluxes"].value)
        assert comparison[key]["wavelengths"].unit == constructed[key]["wavelengths"].unit
        assert comparison[key]["fluxes"].unit == constructed[key]["fluxes"].unit

    cat_base = catfile.split('.')[0]
    outbase = cat_base + '_with_flambda.cat'
    flambda_output_catalog = os.path.join(TEST_DATA_DIR, outbase)
    os.remove(flambda_output_catalog)
    os.remove(sed_catalog)


def test_convert_to_flam():
    """Test basic conversion"""
    magnitudes = np.array([15., 20.])

    photflam = 1e-20
    photfnu = 1e-33
    zeropoint = 25.0
    pivot = 4.44 * u.micron
    params = (photflam, photfnu, zeropoint, pivot)
    mag_sys = 'abmag'
    flam = spec.convert_to_flam(magnitudes, params, mag_sys)
    truth = np.array([5.51426449e-17, 5.51426449e-19])
    assert np.allclose(flam, truth, atol=1e-25)


def test_single_mag_column():
    """Case where input ascii catalog contains only one magntude column.
    In this case extrapolation is necessary."""
    catfile = os.path.join(TEST_DATA_DIR, 'point_sources_one_filter.cat')
    output_hdf5 = os.path.join(TEST_DATA_DIR, 'all_spectra.hdf5')
    sed_catalog = spec.make_all_spectra(catfile, output_filename=output_hdf5)
    comparison = hdf5.open(os.path.join(TEST_DATA_DIR, 'output_spec_from_one_filter.hdf5'))
    constructed = hdf5.open(output_hdf5)
    for key in comparison:
        assert key in constructed.keys()
        assert all(comparison[key]["wavelengths"].value == constructed[key]["wavelengths"].value)
        assert all(comparison[key]["fluxes"].value == constructed[key]["fluxes"].value)
        assert comparison[key]["wavelengths"].unit == constructed[key]["wavelengths"].unit
        assert comparison[key]["fluxes"].unit == constructed[key]["fluxes"].unit

    cat_base = catfile.split('.')[0]
    outbase = cat_base + '_with_flambda.cat'
    flambda_output_catalog = os.path.join(TEST_DATA_DIR, outbase)
    os.remove(flambda_output_catalog)
    os.remove(sed_catalog)


def test_multiple_ascii_catalogs():
    """Case where multiple ascii catalogs are input"""
    catfile = os.path.join(TEST_DATA_DIR, 'point_sources.cat')
    galfile = os.path.join(TEST_DATA_DIR, 'galaxies.cat')
    catalogs = [catfile, galfile]
    output_hdf5 = os.path.join(TEST_DATA_DIR, 'all_spectra.hdf5')
    sed_catalog = spec.make_all_spectra(catalogs, output_filename=output_hdf5)
    comparison = hdf5.open(os.path.join(TEST_DATA_DIR, 'source_sed_file_from_point_sources.hdf5'))
    constructed = hdf5.open(output_hdf5)
    for key in comparison:
        assert key in constructed.keys()
        assert all(comparison[key]["wavelengths"].value == constructed[key]["wavelengths"].value)
        assert all(comparison[key]["fluxes"].value == constructed[key]["fluxes"].value)
        assert comparison[key]["wavelengths"].unit == constructed[key]["wavelengths"].unit
        assert comparison[key]["fluxes"].unit == constructed[key]["fluxes"].unit
    for catfile in catalogs:
        cat_base = catfile.split('.')[0]
        outbase = cat_base + '_with_flambda.cat'
        flambda_output_catalog = os.path.join(TEST_DATA_DIR, outbase)
        os.remove(flambda_output_catalog)
    os.remove(sed_catalog)


def test_spectra_rescaling():
    """Test the functionality for rescaling input spectra to a given
    magnitude in a given filter
    """
    # JWST primary mirror area in cm^2. Needed for countrate check
    # at the end
    primary_area = 25.326 * (u.m * u.m)

    # Create spectrum: one source to be normalized
    # and the other should not be
    waves = np.arange(0.4, 5.6, 0.01)
    flux = np.repeat(1e-16, len(waves))
    flux2 = np.repeat(4.24242424242e-18, len(waves))  # arbitrary value
    spectra = {1: {"wavelengths": waves * u.micron,
                   "fluxes": flux * u.pct},
               2: {"wavelengths": waves * u.micron,
                   "fluxes": flux2 * units.FLAM}}

    # Create source catalog containing scaling info
    catalog = Table()
    catalog['index'] = [1, 2]
    catalog['nircam_f322w2_magnitude'] = [18.] * 2
    catalog['niriss_f090w_magnitude'] = [18.] * 2
    #catalog['fgs_magnitude'] = [18.] * 2

    # Instrument info
    instrument = ['nircam', 'niriss']  # , 'fgs']
    filter_name = ['F322W2', 'F090W']  # , 'N/A']
    module = ['B', 'N']  # , 'F']
    detector = ['NRCA1', 'NIS']  # , 'GUIDER1']

    # Magnitude systems of renormalization magnitudes
    mag_sys = ['vegamag', 'abmag', 'stmag']

    # Loop over options and test each
    for inst, filt, mod, det in zip(instrument, filter_name, module, detector):

        # Extract the appropriate column from the catalog information
        magcol = [col for col in catalog.colnames if inst in col]
        sub_catalog = catalog['index', magcol[0]]

        # Filter throughput files
        filter_thru_file = get_filter_throughput_file(instrument=inst, filter_name=filt,
                                                      nircam_module=mod, fgs_detector=det)

        # Retrieve the correct gain value that goes with the fluxcal info
        if inst == 'nircam':
            gain = MEAN_GAIN_VALUES['nircam']['lwb']
        elif inst == 'niriss':
            gain = MEAN_GAIN_VALUES['niriss']
        elif inst == 'fgs':
            gain = MEAN_GAIN_VALUES['fgs'][det.lower()]

        # Create filter bandpass object, to be used in the final
        # comparison
        filter_tp = ascii.read(filter_thru_file)
        bp_waves = filter_tp['Wavelength_microns'].data * u.micron
        thru = filter_tp['Throughput'].data
        bandpass = SpectralElement(Empirical1D, points=bp_waves, lookup_table=thru) / gain

        # Check the renormalization in all photometric systems
        for magsys in mag_sys:
            rescaled_spectra = spec.rescale_normalized_spectra(spectra, sub_catalog, magsys, filter_thru_file, gain)

            # Calculate the countrate associated with the renormalized
            # spectrum through the requested filter
            for dataset in rescaled_spectra:
                if dataset == 1:
                    # This block is for the spectra that are rescaled
                    rescaled_spectrum = SourceSpectrum(Empirical1D, points=rescaled_spectra[dataset]['wavelengths'],
                                                       lookup_table=rescaled_spectra[dataset]['fluxes'])

                    obs = Observation(rescaled_spectrum, bandpass, binset=bandpass.waveset)
                    renorm_counts = obs.countrate(area=primary_area)

                    # Calculate the countrate associated with an object of
                    # matching magnitude
                    if inst != 'fgs':
                        mag_col = '{}_{}_magnitude'.format(inst.lower(), filt.lower())
                    else:
                        mag_col = 'fgs_magnitude'
                    filt_info = spec.get_filter_info([mag_col], magsys)
                    magnitude = catalog[mag_col][dataset - 1]
                    photflam, photfnu, zeropoint, pivot = filt_info[mag_col]
                    check_counts = magnitude_to_countrate('imaging', magsys, magnitude, photfnu=photfnu.value,
                                                          photflam=photflam.value, vegamag_zeropoint=zeropoint)

                    if magsys != 'vegamag':
                        # As long as the correct gain is used, AB mag and ST mag
                        # count rates agree very well
                        tolerance = 0.0005
                    else:
                        # Vegamag count rates for NIRISS have a sligtly larger
                        # disagreement. Zeropoints were derived assuming Vega = 0.02
                        # magnitudes. This offset has been added to the rescaling
                        # function, but may not be exact.
                        tolerance = 0.0015

                    # This dataset has been rescaled, so check that the
                    # countrate from the rescaled spectrum matches that from
                    # the magnitude it was rescaled to
                    if isinstance(check_counts, u.quantity.Quantity):
                        check_counts = check_counts.value
                    if isinstance(renorm_counts, u.quantity.Quantity):
                        renorm_counts = renorm_counts.value

                    print(inst, filt, magsys, renorm_counts, check_counts, renorm_counts / check_counts)
                    assert np.isclose(renorm_counts, check_counts, atol=0, rtol=tolerance), \
                        print('Failed assertion: ', inst, filt, magsys, renorm_counts, check_counts,
                              renorm_counts / check_counts)
                elif dataset == 2:
                    # Not rescaled. In this case Mirage ignores the magnitude
                    # value in the catalog, so we can't check against check_counts.
                    # Just make sure that the rescaling function did not
                    # change the spectrum at all
                    assert np.all(spectra[dataset]['fluxes'] == rescaled_spectra[dataset]['fluxes'])



"""This module contains functions for working with flux calibration
information, such as filter-based zeropoints, photflam, photfnu, and
pivot wavelegth values
"""
from astropy.io import ascii
from astropy.table import Column
import copy
import numpy as np


def add_detector_to_zeropoints(detector, zeropoint_table):
    """Manually add detector dependence to the zeropoint table for
    NIRCam and NIRISS simualtions. This is being done as a placeholder
    for the future, where we expect zeropoints to be detector-dependent.

    Parameters
    ----------
    detector : str
        Name of detector to add to the table

    zeropoint_table : astropy.table.Table
        Table of filter zeropoint information

    Returns
    -------
    base_table : astropy.table.Table
        Copy of ``zeropoint_table`` with Detector column added
    """
    # Add "Detector" to the list of column names
    base_table = copy.deepcopy(zeropoint_table)
    num_entries = len(zeropoint_table)
    det_column = Column(np.repeat(detector, num_entries), name="Detector")
    base_table.add_column(det_column, index=0)
    return base_table


def fluxcal_info(params, usefilt, detector, module):
    """Retrive basic flux calibration information from the ascii file in
    the repository

    Parameters
    ----------
    params : dict
        Nested dictionary containing the input yaml file parameters

    usefile : str
        Either 'filter' or 'pupil', corresponding to which yaml
        parameter contains the optical element to use for flux cal

    Returns
    -------
    detector : str
        Name of the detector (e.g. NRCA1)
    """
    zpts = read_zeropoint_file(params['Reffiles']['flux_cal'])

    # In the future we expect zeropoints to be detector dependent, as they
    # currently are for FGS. So if we are working with NIRCAM or NIRISS,
    # manually add a Detector key to the dictionary as a placeholder.
    if params["Inst"]["instrument"].lower() in ["nircam", "niriss"]:
        zps = add_detector_to_zeropoints(detector, zpts)
    else:
        zps = copy.deepcopy(zpts)

    # Make sure the requested filter is allowed
    if params['Readout'][usefilt] not in zps['Filter']:
        raise ValueError(("WARNING: requested filter {} is not in the list of "
                          "possible filters.".format(params['Readout'][usefilt])))

    # Get the photflambda and photfnu values that go with
    # the filter
    mtch = ((zps['Detector'] == detector) &
            (zps['Filter'] == params['Readout'][usefilt]) &
            (zps['Module'] == module))
    vegazeropoint = zps['VEGAMAG'][mtch][0]
    photflam = zps['PHOTFLAM'][mtch][0]
    photfnu = zps['PHOTFNU'][mtch][0]
    pivot = zps['Pivot_wave'][mtch][0]

    return vegazeropoint, photflam, photfnu, pivot


def read_zeropoint_file(filename):
    """Read in the ascii table containing all of the flux calibration
    information

    Parameters
    ----------
    filename : str
        Name of ascii file

    Returns
    -------
    flux_table : astropy.table.Table
        Table of flux calibration information
    """
    flux_table = ascii.read(filename)
    return flux_table



"""Based on SIAF transforms PATHS
NIRCAM
------
science --> ideal --> V2,V3
V2,V3 --> ideal --> science
"""
import sys

import numpy as np
from astropy.modeling.models import Polynomial2D, Shift


#t = ascii.read("NIRCam_SIAF_2016-09-29.csv",header_start=1)

def get_siaf_transform(row, aperture, from_system, to_system, degree):
    """
    This reads in the file with transformations that the TEL team
    is using to construct the SIAF file. These transformations are
    defined as two polynomials (one in x, and one in y) transforming coordinates
    from one coordinate system to another.
    Parameters
    ----------
    aperture: str
        Name of aperture on NIRCam, composed of the detector name followed
        by an underscore and the subarray name. (e.g. "NRCA1_FULL",
        "NRCB5_SUB160")
    from_system : str
        Starting system (e.g. "science", "ideal")
    to_system : str
        Ending coordinate system (e.g. "science" , "ideal")
    degree : int
        Degree of polynomial
    Returns
    -------
    x_model : astropy.modeling.Model
        Correction in x
    y_model : astropy.modeling.Model
        Correction in y
    from_units : str
        Units in the starting system
    to_units : str
        Units in the ending system
    Examples
    --------
    >>> get_siaf_transform('NRCA1_FULL', "science", "ideal", 5)
    """
    #read in csv file of coefficients
    #t = ascii.read("NIRCam_SIAF_2016-09-29.csv",header_start=1)
    #t = ascii.read(coefffile,header_start=1)

    #from_system and to_system are very limited. Can only be "ideal" for the
    #distortion-free coords, and "science" for distorted coords
    from_system = from_system.lower()
    if from_system not in ['ideal','science']:
        print("Requested from_system of {} not recognized.".format(from_system))
        sys.exit()

    to_system = to_system.lower()
    if to_system not in ['ideal','science']:
        print("Requested to_system of {} not recognized.".format(to_system))
        sys.exit()

    #Generate the string corresponding to the requested coefficient labels
    if from_system == 'ideal' and to_system == 'science':
        label = 'Idl2Sci'
        from_units = 'arcsec'
        to_units = 'distorted pixels'
    elif from_system == 'science' and to_system == 'ideal':
        label = 'Sci2Idl'
        from_units = 'distorted pixels'
        to_units = 'arcsec'

    #Find the row that matches the requested aperture
    #match = t['AperName'] == aperture
    #if np.any(match) == False:
    #    print("Aperture name {} not found in input CSV file.".format(aperture))
    #    sys.exit()
    #row = t[match]

    #Get the coefficients for "science" to "ideal" transformation (and back)
    #"science" is distorted pixels. "ideal" is undistorted arcsec from the
    #the reference pixel location.

    #we need the parity value, which describes the relationship between the
    #v2,v3 coordinate system and the science pixel coordinate system
    parity = row['VIdlParity'].data[0]

    #Then create the model for the transformation
    X_cols = [c for c in row.colnames if label+'X' in c]
    #for kw in X_cols:
    #    ele = row[kw]
    #    ele *= parity
    #    row[kw] = ele
    x_coeffs = row[X_cols]
    X_model = to_model(x_coeffs,degree)

    Y_cols = [c for c in row.colnames if label+'Y' in c]
    y_coeffs = row[Y_cols]
    Y_model = to_model(y_coeffs,degree)

    return X_model, Y_model, from_units, to_units


def to_model(coeffs, degree=5):
    """
    Creates an astropy.modeling.Model object
    Parameters
    ----------
    coeffs : array like
        Coefficients from the ISIM transformations file.
    degree : int
        Degree of polynomial.
        Default is 5 as in the ISIM file but many of the polynomials are of
        a smaller degree.
    Returns
    -------
    poly : astropy.modeling.Polynomial2D
        Polynomial model transforming one coordinate (x or y) between two systems.
    """

    #map Colin's coefficients into the order expected by Polynomial2D
    c = {}
    for cname in coeffs.colnames:
        siaf_i = int(cname[-2])
        siaf_j = int(cname[-1])
        name = 'c{0}_{1}'.format(siaf_i-siaf_j,siaf_j)
        c[name] = coeffs[cname].data[0]

    #0,0 coefficient should not be used, according to Colin's TR
    #JWST-STScI-001550
    c['c0_0'] = 0

    return Polynomial2D(degree, **c)


def get_siaf_v2v3_transform(row,aperture,from_system='v2v3',to_system='v2v3'):
    """
    Generate transformation model to go to/from V2/V3 from
    undistorted angular distnaces from the reference pixel ("ideal")
    """
    #read in csv file of coefficients
    #t = ascii.read("NIRCam_SIAF_2016-09-29.csv",header_start=1)
    #t = ascii.read(coefffile,header_start=1)

    from_system = from_system.lower()
    to_system = to_system.lower()

    if from_system != 'v2v3' and to_system != 'v2v3':
        print("WARNING, either from_system or to_system must be 'v2v3'")
        sys.exit()

    #Find the row that matches the requested aperture
    #match = t['AperName'] == aperture
    #if np.any(match) == False:
    #    print("Aperture name {} not found in input CSV file.".format(aperture))
    #    sys.exit()
    #row = t[match]

    #Then create the model for the transformation
    parity = row['VIdlParity'].data[0]
    v3_ideal_y_angle = row['V3IdlYAngle'].data[0] * np.pi / 180.

    #print("parity and angle are {}, {}".format(parity,v3_ideal_y_angle))

    X_model, Y_model = v2v3_model(from_system,to_system,parity,v3_ideal_y_angle)

    return X_model, Y_model


def v2v3_model(from_sys, to_sys, par, angle):
    """
    Creates an astropy.modeling.Model object
    for the undistorted ("ideal") to V2V3 coordinate translation
    """
    if from_sys != 'v2v3' and to_sys != 'v2v3':
        print("This function is designed to generate the transformation either to or from V2V3.")
        sys.exit()

    #cast the transform functions as 1st order polynomials
    xc = {}
    yc = {}
    if to_sys == 'v2v3':
        xc['c1_0'] = par * np.cos(angle)
        xc['c0_1'] = np.sin(angle)
        yc['c1_0'] = (0.-par) * np.sin(angle)
        yc['c0_1'] = np.cos(angle)

    if from_sys == 'v2v3':
        xc['c1_0'] = par * np.cos(angle)
        xc['c0_1'] = par * (0. - np.sin(angle))
        yc['c1_0'] = np.sin(angle)
        yc['c0_1'] = np.cos(angle)

    #0,0 coeff should never be used.
    xc['c0_0'] = 0
    yc['c0_0'] = 0

    #print("coeffs for v2v3 transform:")
    #for key in xc:
    #    print("{} {}".format(key,xc[key]))
    #sys.exit()

    xmodel = Polynomial2D(1, **xc)
    ymodel = Polynomial2D(1, **yc)

    return xmodel, ymodel


def get_refpix(row):
    #Return the reference location within the given aperture
    xref = row['XSciRef'].data[0]
    yref = row['YSciRef'].data[0]
    return Shift(-xref) & Shift(-yref)


def get_v2v3ref(row):
    #Return v2 and v3 at the reference location
    #These are arcsec in the SIAF file. Convert to degrees
    v2ref = row['V2Ref'].data[0]
    v3ref = row['V3Ref'].data[0]
    return Shift(v2ref) & Shift(v3ref)



""" Common utilities for the MIRaGe test suite.

Authors
-------
    - Lauren Chambers

"""

import os
import yaml

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))

def parametrized_data():
    """Load parametrized data from file.

    Returns
    -------
    test_data : dict
        Dictionary containing parametrized test data
    """
    parametrized_data_file = os.path.join(__location__, 'test_data', 'parametrized_test_data.yaml')
    with open(parametrized_data_file) as f:
        test_data = yaml.safe_load(f.read())

    return test_data


"""Test the functions provided by siaf_interface.

Authors
-------
    - Johannes Sahlmann

Use
---
    >>> pytest -s test_siaf_interface.py


"""
from pysiaf import iando

from mirage.utils import siaf_interface


def test_sci_subarray_corners():
    """Unit test for siaf_interface.sci_subarray_corners."""
    instrument = 'NIRCam'
    siaf_detector_layout = iando.read.read_siaf_detector_layout()
    master_aperture_names = siaf_detector_layout['AperName'].data
    for aperture_name in master_aperture_names:
        if 'NRC' not in aperture_name:
            continue
        x_sci, y_sci = siaf_interface.sci_subarray_corners(instrument, aperture_name)
        assert len(x_sci) == 2
        assert len(y_sci) == 2



"""Test the functions provided by mirage.psf.deployments

Authors
-------
    - Lauren Chambers

Use
---
    >>> pytest test_deployments.py
"""
import glob
import os
import shutil

import numpy as np
import pytest
import webbpsf

from .utils import parametrized_data
from mirage.psf.deployments import generate_random_ote_deployment, load_ote_from_deployment_yaml
from mirage.utils.utils import ensure_dir_exists

# Define directory and file locations
__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
TEMP_TEST_DIRECTORY = os.path.join(__location__, 'temp_data', 'test_deployments')

# Load parametrized data
PARAMETRIZED_DATA = parametrized_data()['test_deployments']


@pytest.fixture(scope="module")
def test_directory(test_dir=TEMP_TEST_DIRECTORY):
    """Create a test directory.

    Parameters
    ----------
    test_dir : str
        Path to directory used for testing

    Yields
    -------
    test_dir : str
        Path to directory used for testing
    """
    # Create directory and yield its name
    ensure_dir_exists(test_dir)  # creates directory with default mode=511
    yield test_dir

    # Remove directory
    if os.path.isdir(test_dir):
        shutil.rmtree(test_dir)


@pytest.fixture(scope="function")
def remove_yamls_and_fits(test_directory):
    """Deletes YAML and FITS files produced by an OTE deployment function.

    Parameters
    ----------
    test_directory : str
        Path to directory used for testing
    """

    yaml_search_path = os.path.join(test_directory, 'deployment_errors*.yaml')
    yaml_files = glob.glob(yaml_search_path)
    opd_search_path = os.path.join(test_directory, 'OPD*.fits')
    opd_files = glob.glob(opd_search_path)
    yield

    print("Removing OPD yamls and FITS files")
    for file in yaml_files + opd_files:
        os.remove(file)


def test_generate_random_ote_deployment(test_directory, remove_yamls_and_fits):
    """Test the creation of a WebbPSF adjustable OTE object representing a
    perturbed OTE mirror state by randomly generating mirror deployment errors

    Parameters
    ----------
    test_directory : str
        Path to directory used for testing
    remove_yamls_and_fits
        Pytest fixture to clear files after test is run
    """
    # Make the OTE randomly
    ote, segment_tilts, ote_opd_with_tilts = generate_random_ote_deployment(
        test_directory, reduction_factor=0.2, save=True
    )

    # Ensure the OPD yaml and fits files were saved
    yaml_search_path = os.path.join(test_directory, 'deployment_errors*.yaml')
    yaml_files = glob.glob(yaml_search_path)
    opd_search_path = os.path.join(test_directory, 'OPD*.fits')
    opd_files = glob.glob(opd_search_path)
    assert len(yaml_files) >=2, 'Failed to generate randomly deployed OTE'
    assert len(opd_files) >=2, 'Failed to generate randomly deployed OTE'

    # Ensure the resultant OTE is as expected
    assert isinstance(ote, webbpsf.opds.OTE_Linear_Model_WSS), 'Invalid format for OTE.'
    assert ote.shape == (1024, 1024), 'Invalid format for OTE.'
    assert ote.segment_state.shape == (19, 6), 'Invalid format for OTE.'
    check_removed_tip_tilt = ote.segment_state[:-1, :3] == 0.
    assert check_removed_tip_tilt.all(), 'Segment tilts were not removed.'
    check_perturbed = ote.segment_state[:-1, 3:] != 0.
    assert check_perturbed.all(), 'Segment tilts were not applied.'

    # Ensure the segment tilts array is as expected
    assert segment_tilts.shape == (19, 2), 'Invalid format for segment tilts.'
    check_perturbed_tilts = segment_tilts != 0.
    assert check_perturbed_tilts.all(), 'Segment tilts were not applied.'

    # Ensure the OTE array with tilts preserved is as expected
    assert ote_opd_with_tilts.shape == (1024, 1024), 'Invalid format for OTE.'
    check_different_opd = ote_opd_with_tilts == ote.opd
    assert not check_different_opd.all(), 'Segment tilts were not removed.'


def test_load_ote_from_deployment_yaml(test_directory, remove_yamls_and_fits):
    """Test the creation of a WebbPSF adjustable OTE object representing a
    perturbed OTE mirror state by loading from a YAML file.

    Parameters
    ----------
    test_directory : str
        Path to directory used for testing
    remove_yamls_and_fits
        Pytest fixture to clear files after test is run
    """
    # Load known data for comparison
    test_data = PARAMETRIZED_DATA['test_load_ote_from_deployment_yaml']

    # Load the OTE from a YAML file
    deployments_file = os.path.join(__location__, 'test_data', 'deployment_errors_test.yaml')
    ote, segment_tilts, ote_opd_with_tilts = load_ote_from_deployment_yaml(
        deployments_file, test_directory, save=True
    )

    # Ensure the OPD yaml and fits files were saved
    yaml_search_path = os.path.join(test_directory, 'deployment_errors*.yaml')
    yaml_files = glob.glob(yaml_search_path)
    opd_search_path = os.path.join(test_directory, 'OPD*.fits')
    opd_files = glob.glob(opd_search_path)
    assert len(yaml_files) >= 2, 'Failed to generate randomly deployed OTE'
    assert len(opd_files) >= 2, 'Failed to generate randomly deployed OTE'

    # Ensure the resultant OTE is as expected
    assert isinstance(ote, webbpsf.opds.OTE_Linear_Model_WSS), 'Invalid format for OTE.'
    assert ote.shape == (1024, 1024), 'Invalid format for OTE.'
    assert ote.segment_state.shape == (19, 6), 'Invalid format for OTE.'
    np.testing.assert_array_equal(ote.segment_state[:-1, :3], 0.,
                                  err_msg='Segment tilts were not removed.')
    assert not np.array_equiv(ote.segment_state[:-1, 3:], 0.), \
        'Segment tilts were not applied.'
    assert np.isclose(np.mean(ote.opd), 2.1281183145894698e-08), \
        'OTE was not loaded correctly from file.'
    np.testing.assert_allclose(ote.segment_state[:-1, 3:],
                               np.array(test_data['segment_state_no_tilts']),
                               err_msg="OTE was not loaded correctly from file.")


    # Ensure the segment tilts array is as expected
    assert segment_tilts.shape == (19, 2), 'Invalid format for segment tilts.'
    assert not np.array_equiv(segment_tilts, 0.), 'Segment tilts were not applied.'
    np.testing.assert_allclose(segment_tilts,
                               test_data['segment_tilts'],
                               err_msg="OTE was not loaded correctly from file.")

    # Ensure the OTE array with tilts preserved is as expected
    assert ote_opd_with_tilts.shape == (1024, 1024), 'Invalid format for OTE.'
    assert not np.array_equal(ote_opd_with_tilts, ote.opd), 'Segment tilts were not removed.'
    assert np.isclose(np.mean(ote_opd_with_tilts), -1.709744233280506e-05), \
        'OTE was not loaded correctly from file.'



"""Various utility functions.

Authors
-------
    - Lauren Chambers
    - Bryan Hilbert

Use
---
    This module can be imported as such:

    >>> import utils
    siaf_files = get_siaf()

Dependencies
------------
    The user must have a configuration file named ``siaf_config.json``
    placed in mirage/config/ directory.
"""

import copy
import json
import os
import pkg_resources
import re
import yaml

from astropy.io import ascii as asc
import numpy as np
from scipy.stats import sigmaclip

from mirage.utils.constants import CRDS_FILE_TYPES, NIRISS_FILTER_WHEEL_FILTERS, NIRISS_PUPIL_WHEEL_FILTERS


def append_dictionary(base_dictionary, added_dictionary, braid=False):
    """Append the content of added_dictionary key-by-key to the base_dictionary.

    This assumes that the keys refer to lists.

    Parameters
    ----------
    base_dictionary : dict
    added_dictionary : dict
    braid : bool
        If true, the elements of added_dictionary are added in alternating sequence.
        This is used to synchronize parallel observations with the pointing file.

    Returns
    -------
    new_dictionary : dict
        Dictionary where every key holds a list of lists

    """
    new_dictionary = copy.deepcopy(base_dictionary)

    # extract an arbitrary key name
    first_key = [key for i, key in enumerate(base_dictionary.keys()) if i == 0][0]

    # Insert keys from added_dictionary that are not yet present in base_dictionary
    for key in added_dictionary.keys():
        if key not in base_dictionary.keys():
            new_dictionary[key] = ['None'] * len(base_dictionary[first_key])

    # Append the items
    for key in new_dictionary.keys():
        if key not in added_dictionary.keys():
            continue
        # print('{} {}'.format(key, new_dictionary[key]))
        if len(new_dictionary[key]) == 0:
            new_dictionary[key] = added_dictionary[key]
        else:
            if braid:
                # solution from https://stackoverflow.com/questions/3678869/pythonic-way-to-combine-two-lists-in-an-alternating-fashion
                new_dictionary[key] = [sub[i] for i in range(len(added_dictionary[key])) for sub in
                                       [new_dictionary[key], added_dictionary[key]]]
            else:
                new_dictionary[key] = new_dictionary[key] + added_dictionary[key]

    return new_dictionary


def calc_frame_time(instrument, aperture, xdim, ydim, amps):
    """Calculate the readout time for a single frame
    of a given size and number of amplifiers. Note that for
    NIRISS and FGS, the fast readout direction is opposite to
    that in NIRCam, so we switch xdim and ydim so that we can
    keep a single equation.

    Parameters:
    -----------
    instrument : str
        Name of the instrument being simulated

    aperture : str
        Name of aperture being simulated (e.g "NRCA1_FULL")
        Currently this is only used to check for the FGS
        ACQ1 aperture, which uses a unique value of colpad
        below.

    xdim : int
        Number of columns in the frame

    ydim : int
        Number of rows in the frame

    amps : int
        Number of amplifiers used to read out the frame

    Returns:
    --------
    frametime : float
        Readout time in seconds for the frame
    """
    instrument = instrument.lower()
    if instrument == "nircam":
        colpad = 12

        # Fullframe
        if amps == 4:
            rowpad = 1
            fullpad = 1
        else:
            # All subarrays
            rowpad = 2
            fullpad = 0

            if ((xdim <= 8) & (ydim <= 8)):
                # The smallest subarray
                rowpad = 3

    elif instrument == "niriss":
        # Reverse x and y since NIRISS's fast readout direction is
        # opposite of NIRCam's
        tmpx = copy.deepcopy(xdim)
        xdim = copy.deepcopy(ydim)
        ydim = tmpx

        colpad = 12

        # Fullframe
        if amps == 4:
            rowpad = 1
            fullpad = 1
        else:
            rowpad = 2
            fullpad = 0

    elif instrument == 'fgs':
        # Reverse x and y since FGS's fast readout direction is
        # opposite of NIRCam's
        tmpx = copy.deepcopy(xdim)
        xdim = copy.deepcopy(ydim)
        ydim = tmpx

        colpad = 12
        fullpad = 0

        if ((xdim == 2048) & (ydim == 2048)):
            rowpad = 1
            fullpad = 1
        else:
            rowpad = 2

        if ((xdim <= 32) & (ydim <= 32)):
            colpad = 6
            rowpad = 1

    return ((1.0 * xdim / amps + colpad) * (ydim + rowpad) + fullpad) * 1.e-5


def check_niriss_filter(oldfilter, oldpupil):
    """
    This is a utility function that checks the FILTER and PUPIL parameters read in from the .yaml file and makes sure
    that for NIRISS the filter and pupil names are correct, releaving the user of the need to remember which of the 12
    filters are in the filter wheel and which are in the pupil wheel.  If the user puts the correct values for filter and
    pupil nothing is done, but the user can just put the filter name in the filter parameter and CLEAR in the pupil
    parameter, in which case this routine sorts out which value actually goes where.  Hence for example one can have a
    parameter file with FILTER = F277W and PUPIL = CLEAR or FILTER = F090W and PUPIL = CLEAR for imaging and either will
    produce the desired result although the actual pairings would be F277W/CLEARP and CLEAR/F090W respectively.

    The code also corrects CLEARP to CLEAR if needed.

    Parameters
    ----------

    oldfilter :  str
        Assumed to be the self.params['Readout']['filter'] value from the .yaml file

    oldpupil :  str
        Assumed to be the self.params['Readout']['pupil'] value from the .yaml file

    Returns
    -------

    newfilter : str
        The proper FILTER parameter for the requested NIRISS filter name

    newpupil : str
        The proper PUPIL parameter for the requested NIRISS filter name

    Note that this routine should only be called when the instrument is NIRISS.
    """
    str1 = oldfilter
    str2 = oldpupil

    if oldfilter in NIRISS_PUPIL_WHEEL_FILTERS:
        newfilter = str2
        newpupil = str1
        if newfilter == 'CLEARP':
            newfilter = 'CLEAR'
    elif oldfilter in NIRISS_FILTER_WHEEL_FILTERS:
        if oldpupil == 'CLEAR':
            newpupil = 'CLEARP'
        else:
            newpupil = oldpupil
        newfilter = oldfilter
    else:
        # Case where oldfilter is an optic other than a filter (e.g. a grism)
        newfilter = oldfilter
        newpupil = oldpupil
    return newfilter, newpupil


def crop_to_subarray(data, bounds):
    """
    Crop the given full frame array down to the appropriate
    subarray size and location based on the requested subarray
    name.

    Parameters
    ----------
    data : numpy.ndarray
        Full frame image or ramp. (x,y) = (2048, 2048)
        May be 2D, 3D, or 4D

    bounds : list
        4-element list containing the full frame indices that
        define the position of the subarray.
        [xstart, ystart, xend, yend]

    Returns
    -------
    data : numpy.ndarray
        Input array cropped in x and y dimensions
    """
    dimensions = len(data.shape)
    yl, xl = data.shape[-2:]

    valid = [False, False, False, False]
    valid = [(b >= 0 and b < xl) for b in bounds[0:3:2]]
    validy = [(b >= 0 and b < yl) for b in bounds[1:4:2]]
    valid.extend(validy)

    if all(valid):
        if dimensions == 2:
            return data[bounds[1]:bounds[3] + 1, bounds[0]:bounds[2] + 1]
        elif dimensions == 3:
            return data[:, bounds[1]:bounds[3] + 1, bounds[0]:bounds[2] + 1]
        elif dimensions == 4:
            return data[:, :, bounds[1]:bounds[3] + 1, bounds[0]:bounds[2] + 1]
        else:
            raise ValueError(("In crop_to_subarray, input array is not 2, 3, or 4D."))
    else:
            raise ValueError(("WARNING: subarray bounds are outside the "
                              "dimensions of the input array."))


def ensure_dir_exists(fullpath):
    """Creates dirs from ``fullpath`` if they do not already exist.
    """
    if not os.path.exists(fullpath):
        os.makedirs(fullpath)


def expand_environment_variable(variable_name, offline=False):
    """ Expand an environment variable and check that the directory exists

    Parameters
    ----------
    variable_name : str
        Environment variable name

    Returns
    -------
    variable_directory : str
        Path pointed to by the environment variable
    """
    variable_directory = os.environ.get(variable_name)
    if variable_directory is None:
        raise ValueError(("{} environment variable is not set. "
                          "This must be set to the base directory "
                          "containing the darks, cosmic ray, PSF, etc "
                          "input files needed for the simulation. "
                          "These files must be downloaded separately "
                          "from the Mirage package.".format(variable_name)))
    if not offline:
        if not os.path.isdir(variable_directory):
            raise FileNotFoundError(("The directory contained in the {} "
                                     "environment variable: {} does not exist or "
                                     "is not accessible.".format(variable_name, variable_directory)))
    return variable_directory


def full_paths(params, module_path, crds_dictionary, offline=False):
    """Expand the relevant input paths from the input yaml file to be full
    paths.

    Parameters
    ----------
    params : dict
        Nested dictionary read in from an input yaml file

    module_path : str
        Path to the Mirage module

    crds_dictionary : dict
        Dictionary of basic observation metadata that is used by CRDS to
        select appropriate reference files.

    offline : bool
        Set to True in order to skip downloading reference files from CRDS
        and instead only return the dictionary of reference file names.

    Returns
    -------
    params : dict
        Modified nested dictionary including full paths
    """
    # Place import statement here in order to avoid circular import
    # with crd_tools
    from mirage.reference_files import crds_tools
    pathdict = {'Reffiles': ['dark', 'linearized_darkfile', 'badpixmask',
                             'superbias', 'linearity', 'saturation', 'gain',
                             'pixelflat', 'illumflat', 'ipc', 'astrometric',
                             'crosstalk', 'occult', 'pixelAreaMap',
                             'subarray_defs', 'filtpupilcombo',
                             'flux_cal', 'readpattdefs', 'filter_throughput',
                             'filter_wheel_positions'],
                'simSignals': ['pointsource', 'psfpath', 'galaxyListFile',
                               'extended', 'movingTargetList', 'movingTargetSersic',
                               'movingTargetExtended', 'movingTargetToTrack',
                               'psf_wing_threshold_file', 'zodiacal', 'scattered'],
                'cosmicRay': ['path'],
                'newRamp': ['dq_configfile', 'sat_configfile', 'superbias_configfile',
                            'refpix_configfile', 'linear_configfile'],
                'Output': ['file', 'directory']}

    all_config_files = {'nircam': {'Reffiles-subarray_defs': 'NIRCam_subarray_definitions.list',
                                   'Reffiles-flux_cal': 'NIRCam_zeropoints.list',
                                   'Reffiles-crosstalk': 'xtalk20150303g0.errorcut.txt',
                                   'Reffiles-readpattdefs': 'nircam_read_pattern_definitions.list',
                                   'Reffiles-filter_throughput': 'placeholder.txt',
                                   'Reffiles-filtpupilcombo': 'nircam_filter_pupil_pairings.list',
                                   'Reffiles-filter_wheel_positions': 'nircam_filter_and_pupil_wheel_positions.txt',
                                   'simSignals-psf_wing_threshold_file': 'nircam_psf_wing_rate_thresholds.txt',
                                   'newRamp-dq_configfile': 'dq_init.cfg',
                                   'newRamp-sat_configfile': 'saturation.cfg',
                                   'newRamp-superbias_configfile': 'superbias.cfg',
                                   'newRamp-refpix_configfile': 'refpix.cfg',
                                   'newRamp-linear_configfile': 'linearity.cfg'},
                        'niriss': {'Reffiles-subarray_defs': 'niriss_subarrays.list',
                                   'Reffiles-flux_cal': 'niriss_zeropoints.list',
                                   'Reffiles-crosstalk': 'niriss_xtalk_zeros.txt',
                                   'Reffiles-readpattdefs': 'niriss_readout_pattern.txt',
                                   'Reffiles-filter_throughput': 'placeholder.txt',
                                   'Reffiles-filtpupilcombo': 'niriss_dual_wheel_list.txt',
                                   'Reffiles-filter_wheel_positions': 'niriss_filter_and_pupil_wheel_positions.txt',
                                   'simSignals-psf_wing_threshold_file': 'niriss_psf_wing_rate_thresholds.txt',
                                   'newRamp-dq_configfile': 'dq_init.cfg',
                                   'newRamp-sat_configfile': 'saturation.cfg',
                                   'newRamp-superbias_configfile': 'superbias.cfg',
                                   'newRamp-refpix_configfile': 'refpix.cfg',
                                   'newRamp-linear_configfile': 'linearity.cfg'},
                        'fgs': {'Reffiles-subarray_defs': 'guider_subarrays.list',
                                'Reffiles-flux_cal': 'guider_zeropoints.list',
                                'Reffiles-crosstalk': 'guider_xtalk_zeros.txt',
                                'Reffiles-readpattdefs': 'guider_readout_pattern.txt',
                                'Reffiles-filter_throughput': 'placeholder.txt',
                                'Reffiles-filtpupilcombo': 'guider_filter_dummy.list',
                                'Reffiles-filter_wheel_positions': 'dummy.txt',
                                'simSignals-psf_wing_threshold_file': 'fgs_psf_wing_rate_thresholds.txt',
                                'newRamp-dq_configfile': 'dq_init.cfg',
                                'newRamp-sat_configfile': 'saturation.cfg',
                                'newRamp-superbias_configfile': 'superbias.cfg',
                                'newRamp-refpix_configfile': 'refpix.cfg',
                                'newRamp-linear_configfile': 'linearity.cfg'}}
    config_files = all_config_files[params['Inst']['instrument'].lower()]

    for key1 in pathdict:
        for key2 in pathdict[key1]:
            if params[key1][key2].lower() not in ['none', 'config', 'crds']:
                params[key1][key2] = os.path.abspath(os.path.expandvars(params[key1][key2]))
            elif params[key1][key2].lower() == 'config':
                cfile = config_files['{}-{}'.format(key1, key2)]
                fpath = os.path.join(module_path, 'config', cfile)
                params[key1][key2] = fpath
                print("'config' specified: Using {} for {}:{} input file".format(fpath, key1, key2))
            elif key2 in CRDS_FILE_TYPES.keys() and params[key1][key2].lower() == 'crds':
                # 'crds' set for one of the reference files means to
                # search for the best reference file currently in CRDS
                # and download if it is not already present in
                # self.crds_datadir
                mapping = crds_tools.get_reffiles(crds_dictionary, [CRDS_FILE_TYPES[key2]], download=not offline)
                params[key1][key2] = mapping[CRDS_FILE_TYPES[key2]]
                print("From CRDS, found {} as the {} reference file.".format(mapping[CRDS_FILE_TYPES[key2]], key2))

                # If we grab the IPC reference file from CRDS, then we need to invert it prior to use
                if key2 == 'ipc':
                    ipc_path, ipc_file = os.path.split(mapping['ipc'])
                    inverted_kernel_file = os.path.join(ipc_path, 'Kernel_to_add_IPC_effects_from_{}'.format(ipc_file))
                    # Check to see if the version of the IPC file with an
                    # inverted kernel already exists.
                    if os.path.isfile(inverted_kernel_file):
                        print("Found an existing inverted kernel for this IPC file: {}".format(inverted_kernel_file))
                        params[key1][key2] = inverted_kernel_file
                        params['Reffiles']['invertIPC'] = False
                    else:
                        print("No inverted IPC kernel file found. Kernel will be inverted prior to use.")
                        params['Reffiles']['invertIPC'] = True
    return params


def get_all_reffiles(param_dict):
    """
    """
    # Place import statement here in order to avoid circular import
    # with crd_tools
    from mirage.reference_files import crds_tools

    mapping = crds_tools.get_reffiles(param_dict, list(CRDS_FILE_TYPES.values()))
    return mapping


def get_siaf():
    '''Return a dictionary that holds the contents of the SIAF config
    file.

    Returns
    -------
    siaf_files : dict
        A dictionary that holds the contents of the config file.
    '''
    utils_dir = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
    package_dir = os.path.dirname(utils_dir)
    config_file = os.path.join(package_dir, 'config', 'siaf_config.json')

    with open(config_file, 'r') as config_file:
        siaf_files = json.load(config_file)

    return siaf_files


def get_aperture_definition(aperture_name, instrument):
    '''Parses the SIAF to get the definition of a given aperture
    '''

    siaf_file = get_siaf()[instrument]
    siaf_table = asc.read(siaf_file, header_start=1)
    row_ind = list(siaf_table['AperName']).index(aperture_name)
    siaf_row = siaf_table[row_ind]

    detector = aperture_name[3:5]  # Won't work for ALL, A, or B case

    if re.search(r"_F\d\d\d[MWN]", aperture_name):
        filt = aperture_name.split('_')[-1]
    else:
        filt = 'ANY'

    refpix_x = siaf_row['XSciRef']
    refpix_y = siaf_row['YSciRef']
    refpix_v2 = siaf_row['V2Ref']
    refpix_v3 = siaf_row['V3Ref']

    x_size = siaf_row['XSciSize']
    y_size = siaf_row['YSciSize']

    xstart = refpix_x - x_size / 2
    xend = refpix_x + x_size / 2
    ystart = refpix_y - y_size / 2
    yend = refpix_y + y_size / 2

    # Need to add num_amps

    aperture_definition = [aperture_name, detector, filt, xstart, ystart, xend,
                           yend, refpix_x, refpix_y, refpix_v2, refpix_v3]
    return aperture_definition


def get_frame_count_info(numints, numgroups, numframes, numskips, numresets):
    """Calculate information on the number of frames per group and
    per integration

    Parameters
    ----------
    numints : int
        Number of integraitons in the exposure

    numgroups : int
        Number of groups per integration

    numframes : int
        Number of frames averaged together to create a group

    numskips : int
        Number of skipped frames per group

    numresets : int
        Number of detector resets between integrations

    Returns
    -------
    frames_per_group
    """
    frames_per_group = numframes + numskips
    frames_per_integration = numgroups * frames_per_group
    total_frames = numgroups * frames_per_group

    if numints > 1:
        # Frames for all integrations
        total_frames *= numints
        # Add the resets for all but the first integration
        total_frames += (numresets * (numints - 1))

    return frames_per_group, frames_per_integration, total_frames


def get_filter_throughput_file(instrument, filter_name, nircam_module=None, fgs_detector=None):
    """Locate the filter throughput file in the config directory that
    corresponds to the given instrument/module/filter

    Parameters
    ----------
    instrument : str
        Name of JWST instrument.

    filter_name : str
        Name of the filter. Ignored in the case of FGS.

    nircam_module : str
        Name of module. Ignored except in the case of NIRCam

    fgs_detector : str
        Name of FGS detector (e.g. 'GUIDER1'). Ignored in the case of other
        instruments

    Returns
    -------
    throughput_file : str
        Name of ascii file containing the filter throughput curve
    """
    modpath = pkg_resources.resource_filename('mirage', '')
    config_path = os.path.join(modpath, 'config')

    instrument = instrument.lower()
    if instrument == 'nircam':
        tp_file = '{}_nircam_plus_ote_throughput_mod{}_sorted.txt'.format(filter_name.upper(),
                                                                          nircam_module.lower())
    elif instrument == 'niriss':
        tp_file = '{}_niriss_throughput1.txt'.format(filter_name.lower())
    elif instrument == 'fgs':
        det_number = fgs_detector[-1]
        tp_file = 'guider{}_throughput_py.txt'.format(det_number)
    else:
        raise ValueError('ERROR: invalid instrument: {}'.format(instrument))

    throughput_file = os.path.join(config_path, tp_file)
    return throughput_file


def get_subarray_info(params, subarray_table):
    """Find aperture-specific information from the subarray information config file

    Parameters:
    -----------
    params : dict
        Nested dictionary containing Mirage parameters.

    subarray_table : astropy.table.Table
        Table containing subarray definition information. Output from
        read_subarray_definition_file

    Returns:
    --------
    params : dict
        Updated nested dictionary
    """
    if params['Readout']['array_name'] in subarray_table['AperName']:
        mtch = params['Readout']['array_name'] == subarray_table['AperName']
        namps = subarray_table['num_amps'].data[mtch][0]
        if namps != 0:
            params['Readout']['namp'] = int(namps)
        else:
            try:
                if ((params['Readout']['namp'] == 1) or
                   (params['Readout']['namp'] == 4)):
                    print(("CAUTION: Aperture {} can be used with either "
                           "a 1-amp".format(subarray_table['AperName'].data[mtch][0])))
                    print("or a 4-amp readout. The difference is a factor of 4 in")
                    print(("readout time. You have requested {} amps."
                           .format(params['Readout']['namp'])))
                else:
                    raise ValueError(("WARNING: {} requires the number of amps to be 1 or 4. Please set "
                                      "'Readout':'namp' in the input yaml file to one of these values."
                                      .format(params['Readout']['array_name'])))
            except KeyError:
                raise KeyError(("WARNING: 'Readout':'namp' not present in input yaml file. "
                                "{} aperture requires the number of amps to be 1 or 4. Please set "
                                "'Readout':'namp' in the input yaml file to one of these values."
                                .format(params['Readout']['array_name'])))
    else:
        raise ValueError(("WARNING: subarray name {} not found in the "
                          "subarray dictionary {}."
                          .format(params['Readout']['array_name'],
                                  params['Reffiles']['subarray_defs'])))
    return params


def magnitude_to_countrate(observation_mode, magsys, mag, photfnu=None, photflam=None,
                           vegamag_zeropoint=None):
    """Convert a given source magnitude into count rate

    Parameters
    ----------
    observation_mode : str
        e.g. 'imaging', 'wfss'

    magsys : str
        Magnitude system of the input magnitudes. Allowed values are:
        'abmag', 'stmag', 'vegamag'

    mag : float or list
        Magnitude value(s) to transform

    photfnu : float
        Photfnu value that relates count rate and flux density. Only used
        in ABMAG conversions

    photflam : float
        Photflam value that relates count rate and flux density. Only used
        in STMAG conversions

    vegamag_zeropoint : float
        Filter offset in the VEGAMAG (or A0V mag) system. Only used in
        VEGAMAG conversions.


    Returns
    -------
    count_rate : float or list
        Count rate (e/s) corresponding to the input magnutude(s)

    """
    # For NIRISS AMI mode, the count rate values calculated need to be
    # scaled by a factor 0.15/0.84 = 0.17857.  The 0.15 value is the
    # throughput of the NRM, while the 0.84 value is the throughput of the
    # imaging CLEARP element that is in place in the pupil wheel for the
    # normal imaging observations.
    if observation_mode in ['ami', 'AMI']:
        count_scale = 0.15 / 0.84
    else:
        count_scale = 1.
    if magsys.lower() == 'abmag':
        try:
            return count_scale * (10**((mag + 48.599934378) / -2.5) / photfnu)
        except:
            raise ValueError(("AB mag to countrate conversion failed."
                              "magnitude = {}, photfnu = {}".format(mag, photfnu)))
    if magsys.lower() == 'vegamag':
        try:
            return count_scale * (10**((vegamag_zeropoint - mag) / 2.5))
        except:
            raise ValueError(("Vega mag to countrate conversion failed."
                              "magnitude = {}".format(mag)))
    if magsys.lower() == 'stmag':
        try:
            return count_scale * (10**((mag + 21.099934378) / -2.5) / photflam)
        except:
            raise ValueError(("ST mag to countrate conversion failed."
                              "magnitude = {}, photflam = {}".format(mag, photflam)))


def sigma_clipped_mean_value_of_image(array, sigma_value):
    """Return the sigma-clipped mean value of a 2D array

    Parameters
    ----------
    array : numpy.ndarray
        array of values

    sigma_value : int
        Number of sigma to use as clipping threshold. Clipping will be
        symmetric above and below the mean

    Returns
    -------
    meanval : float
        3-sigma iteratively sigma-clipped mean value of ``array``
    """
    good = np.isfinite(array)
    clipped, lo, hi = sigmaclip(array[good], low=sigma_value, high=sigma_value)
    meanval = clipped.mean()
    return meanval


def parse_RA_Dec(ra_string, dec_string):
    """Convert input RA and Dec strings to floats

    Parameters
    ----------

    ra_string : str
        String containing RA information. Can be in the form 10:12:13.2
        or 10h12m13.2s

    dec_string : str
        String containing Declination information. Can be in the form
        10:12:2.4 or 10d12m2.4s

    Returns
    -------

    ra_degrees : float
        Right Ascention value in degrees

    dec_degrees : float
        Declination value in degrees
    """
    try:
        ra_string = ra_string.lower()
        ra_string = ra_string.replace("h", ":")
        ra_string = ra_string.replace("m", ":")
        ra_string = ra_string.replace("s", "")
        ra_string = re.sub(r"\s+", "", ra_string)
        dec_string = dec_string.lower()
        dec_string = dec_string.replace("d", ":")
        dec_string = dec_string.replace("m", ":")
        dec_string = dec_string.replace("s", "")
        dec_string = re.sub(r"\s+", "", dec_string)

        values = ra_string.split(":")
        ra_degrees = 15.*(int(values[0]) + int(values[1])/60. + float(values[2])/3600.)

        values = dec_string.split(":")
        if "-" in values[0]:
            sign = -1
            values[0] = values[0].replace("-", "")
        else:
            sign = +1

        dec_degrees = sign*(int(values[0]) + int(values[1])/60. + float(values[2])/3600.)
        return ra_degrees, dec_degrees
    except:
        raise ValueError("Error parsing RA, Dec strings: {} {}".format(ra_string, dec_string))


def read_pattern_check(parameters):
    # Check the readout pattern that's entered and set nframe and nskip
    # accordingly
    parameters['Readout']['readpatt'] = parameters['Readout']['readpatt'].upper()

    # Read in readout pattern definition file
    # and make sure the possible readout patterns are in upper case
    readpatterns = asc.read(parameters['Reffiles']['readpattdefs'])
    readpatterns['name'] = [s.upper() for s in readpatterns['name']]

    # If the requested readout pattern is in the table of options,
    # then adopt the appropriate nframe and nskip
    if parameters['Readout']['readpatt'] in readpatterns['name']:
        mtch = parameters['Readout']['readpatt'] == readpatterns['name']
        parameters['Readout']['nframe'] = int(readpatterns['nframe'][mtch].data[0])
        parameters['Readout']['nskip'] = int(readpatterns['nskip'][mtch].data[0])
        print(('Requested readout pattern {} is valid. '
               'Using the nframe = {} and nskip = {}'
               .format(parameters['Readout']['readpatt'],
                       parameters['Readout']['nframe'],
                       parameters['Readout']['nskip'])))

        print('in read_pattern_check, nframe and nskip are:')
        print(parameters['Readout']['nframe'], parameters['Readout']['nskip'])
        print(type(parameters['Readout']['nframe']), type(parameters['Readout']['nskip']))
        print('maxiter', parameters['nonlin']['maxiter'], type(parameters['nonlin']['maxiter']))

    else:
        # If the read pattern is not present in the definition file
        # then quit.
        raise ValueError(("WARNING: the {} readout pattern is not defined in {}."
                          .format(parameters['Readout']['readpatt'],
                                  parameters['Reffiles']['readpattdefs'])))
    return parameters


def read_subarray_definition_file(filename):
    """Read in the file that contains a list of subarray names and related information

    Parameters:
    -----------
    filename : str
        Name of the ascii file containing the table of subarray information

    Returns:
    --------
    data : astropy.table.Table
        Table containing subarray information
    """
    try:
        data = asc.read(filename, data_start=1, header_start=0)
    except:
        raise RuntimeError(("Error: could not read in subarray definitions file: {}"
                            .format(filename)))
    return data


def read_yaml(filename):
    """Read the contents of a yaml file into a nested dictionary

    Parameters
    ----------
    filename : str
        Name of yaml file to be read in

    Returns
    -------
    data : dict
        Nested dictionary of file contents
    """
    try:
        with open(filename, 'r') as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
    except FileNotFoundError as e:
            print(e)
    return data


def write_yaml(data, filename):
    """Write a nested dictionary to a yaml file

    Parameters
    ----------
    data : dict
        Nested dictionary of paramters

    filename : str
        Output filename
    """
    with open(filename, 'w') as output:
        yaml.dump(data, output, default_flow_style=False)



#! /usr/bin/env python

"""
This module contains functions used to indentify and download reference files
from CRDS and place them in the expected location for Mirage reference files.

This module uses the crds software package
(https://hst-crds.stsci.edu/static/users_guide/index.html) which is installed when
the JWST calibration pipeline package is installed. Reference files are
identified by supplying some basic metadata from the exposure being calibrated.

See https://hst-crds.stsci.edu/static/users_guide/library_use.html#crds-getreferences
for a description of the function used for this task.

Author
------

    - Bryan Hilbert


Use
---

    This module can be used as such:
    ::
        from mirage.reference_files import crds
        params = {'INSTRUME': 'NIRCAM', 'DETECTOR': 'NRCA1'}
        reffiles = crds.get_reffiles(params)
"""

import datetime
import os

from mirage.utils.utils import ensure_dir_exists
from mirage.utils.constants import EXPTYPES


def env_variables():
    """Check the values of the CRDS-related environment variables

    Returns
    ----------
    crds_data_path : str
        Full path to the location of the CRDS reference files
    """
    crds_data_path = path_check()
    server_check()
    return crds_data_path


def path_check():
    """Check that the CRDS_PATH environment variable is set. This will be
    the location to which CRDS reference files are downloaded. If the env
    variable is not set, default to use $HOME/crds_cache/

    Returns
    -------
    crds_path : str
        Full path to the location of the CRDS reference files
    """
    crds_path = os.environ.get('CRDS_PATH')
    if crds_path is None:
        reffile_dir = '{}/crds_cache'.format(os.environ.get('HOME'))
        os.environ["CRDS_PATH"] = reffile_dir
        ensure_dir_exists(reffile_dir)
        print('CRDS_PATH environment variable not set. Setting to {}'.format(reffile_dir))
        return reffile_dir
    else:
        return crds_path


def server_check():
    """Check that the CRDS_SERVER_URL environment variable is set. This
    controls where Mirage will look for CRDS information. If the env
    variable is not set, set it to the JWST CRDS server.
    """
    crds_server = os.environ.get('CRDS_SERVER_URL')
    if crds_server is None:
        os.environ["CRDS_SERVER_URL"] = "https://jwst-crds.stsci.edu"


def dict_from_yaml(yaml_dict):
    """Create a dictionary to be used as input to the CRDS getreferences
    function from the nested dictionary created when a standard Mirage
    input yaml file is read in.

    Parameters
    ----------
    yaml_dict : dict
        Nested dictionary from reading in yaml file

    Returns
    -------
    crds_dict : dict
        Dictionary of information necessary to select refernce files
        via getreferences().
    """
    crds_dict = {}
    instrument = yaml_dict['Inst']['instrument'].upper()
    crds_dict['INSTRUME'] = instrument
    crds_dict['READPATT'] = yaml_dict['Readout']['readpatt'].upper()

    # Currently, all reference files that use SUBARRAY as a selection
    # criteria contain SUBARRAY = 'GENERIC', meaning that SUBARRAY
    # actually isn't important. So let's just set it to FULL here.
    crds_dict['SUBARRAY'] = 'FULL'

    # Use the current date and time in order to get the most recent
    # reference file
    crds_dict['DATE-OBS'] = datetime.date.today().isoformat()
    current_date = datetime.datetime.now()
    crds_dict['TIME-OBS'] = current_date.time().isoformat()

    array_name = yaml_dict['Readout']['array_name']
    crds_dict['DETECTOR'] = array_name.split('_')[0].upper()
    if '5' in crds_dict['DETECTOR']:
        crds_dict['DETECTOR'] = crds_dict['DETECTOR'].replace('5', 'LONG')

    if 'FGS' in crds_dict['DETECTOR']:
        crds_dict['DETECTOR'] = 'GUIDER{}'.format(crds_dict['DETECTOR'][-1])

    if instrument == 'NIRCAM':
        if crds_dict['DETECTOR'] in ['NRCALONG', 'NRCBLONG']:
            crds_dict['CHANNEL'] = 'LONG'
        else:
            crds_dict['CHANNEL'] = 'SHORT'

    # For the purposes of choosing reference files, the exposure type should
    # always be set to imaging, since it is used to locate sources in the
    # seed image, prior to any dispersion.
    crds_dict['EXP_TYPE'] = EXPTYPES[instrument.lower()]["imaging"]

    # This assumes that filter and pupil names match up with reality,
    # as opposed to the more user-friendly scheme of allowing any
    # filter to be in the filter field.
    crds_dict['FILTER'] = yaml_dict['Readout']['filter']
    crds_dict['PUPIL'] = yaml_dict['Readout']['pupil']
    return crds_dict


def get_reffiles(parameter_dict, reffile_types, download=True):
    """Determine CRDS's best reference files to use for a particular
    observation, and download them if they are not already present in
    the CRDS_PATH. The determination is made based on the information in
    the parameter_dictionary.

    Parameters
    ----------
    parameter_dict : dict
        Dictionary of basic metadata from the file to be processed by the
        returned reference files (e.g. INSTRUME, DETECTOR, etc)

    reffile_types : list
        List of reference file types to look up and download. These must
        be contained in CRDS's list of reference file types.

    download : bool
        If True (default), the identified best reference files will be
        downloaded. If False, the dictionary of best reference files will
        still be returned, but the files will not be downloaded. The use
        of False is primarily intended to support testing on Travis.
    """
    # IMPORTANT: Import of crds package must be done AFTER the environment
    # variables are set in the functions above
    import crds
    from crds import CrdsLookupError

    if download:
        try:
            reffile_mapping = crds.getreferences(parameter_dict, reftypes=reffile_types)
        except CrdsLookupError as e:
            raise ValueError("ERROR: CRDSLookupError when trying to find reference files for parameters: {}".format(parameter_dict))

        # Check for missing files
        for key, value in reffile_mapping.items():
            if "NOT FOUND" in value:
                raise ValueError("ERROR: No {} reference file found when using parameter dictionary: {}".format(key, parameter_dict))
    else:
        # If the files will not be downloaded, still return the same local
        # paths that are returned when the files are downloaded. Note that
        # this follows the directory structure currently assumed by CRDS.
        crds_path = os.environ.get('CRDS_PATH')
        try:
            reffile_mapping = crds.getrecommendations(parameter_dict, reftypes=reffile_types)
        except CrdsLookupError as e:
            raise ValueError("ERROR: CRDSLookupError when trying to find reference files for parameters: {}".format(parameter_dict))
        for key, value in reffile_mapping.items():

            # Check for NOT FOUND must be done here because the following
            # line will raise an exception if NOT FOUND is present
            if "NOT FOUND" in value:
                raise ValueError("ERROR: No {} reference file found when using parameter dictionary: {}".format(key, parameter_dict))
            instrument = value.split('_')[1]
            reffile_mapping[key] = os.path.join(crds_path, 'references/jwst', instrument, value)

    return reffile_mapping



#! /usr/bin/env python

"""Catalog-related photometric tools

e.g. Translate magnitudes from one filter to another

"""

import astropy.units as u


def translate(input_spectrum, magnitude, band, new_band, renormalize=False):
    """Calculate a source's magnitude in a given filter given the magnitude in
    a different band and a spectrum.

    If renormalize is False, then the spectrum is will be used directly.
    If renormalize is True, then the input spectrum will be renormalized
    such that it produces the given magnitude in 'band'. The magnitude in
    'new_band" is then calculated.

    Parameters
    ----------
    input_spectrum : ??

    magnitude : float
        astropy magnitude unit attached?

    band : str
        Name of filter passband corresponding to 'magnitude'
        (This implies Mirage knows about the magnitude band by name. What about
        a case where the user supplies the filter passband?)

    new_band : str
        Name of band to translate source's brightness into. Again, this assumes mirage
        knows the name of the filter. What about a case where the user provides a filter
        bandpass?

    renormalize : bool
        If False, the input spectrum is used directly and 'magnitude' is ignored.
    """
    something
    return new_magnitude




import re
import sys
from pkg_resources import get_distribution, DistributionNotFound

__version_commit__ = ''
_regex_git_hash = re.compile(r'.*\+g(\w+)')

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    __version__ = 'dev'

if '+' in __version__:
    commit = _regex_git_hash.match(__version__).groups()
    if commit:
        __version_commit__ = commit[0]



"""Unit tests for flux calibration/zeropoints

Authors
-------
    - Bryan Hilbert

Use
---
    Ensure you have pytest installed. Then, simply run pytest in any
    parent directory of mirage/tests/:
    >>> pytest
"""
from astropy.table import Table
import numpy as np
import os
import pkg_resources

from mirage.utils import flux_cal

package_path = pkg_resources.resource_filename('mirage', '')
CONFIG_DIR = os.path.join(package_path, 'config')


def test_add_detector_to_zeropoints():
    """Test addition of column to table
    """
    detector = 'NRCA1'
    tab = Table()
    tab['index'] = np.arange(5)
    tab['information'] = [1.2, 2.3, 3.4, 4.5, 5.6]

    updated_tab = flux_cal.add_detector_to_zeropoints(detector, tab)
    assert np.all(updated_tab['Detector'].data == np.array([detector] * 5))
    assert np.all(updated_tab['index'].data == tab['index'].data)
    assert np.all(updated_tab['information'].data == tab['information'].data)


def test_fluxcal_info():
    """Test that zeropoint information for the exposure is correctly retrieved
    """
    params = {'Inst': {"instrument": 'NIRCAM'},
              'Readout': {'filter': 'F200W', 'pupil': 'CLEAR'},
              'Reffiles': {'flux_cal': os.path.join(CONFIG_DIR, 'NIRCam_zeropoints.list')}
              }

    usefilt = 'filter'
    detector = 'NRCA1'
    module = 'A'
    vegazp, photflam, photfnu, pivot = flux_cal.fluxcal_info(params, usefilt, detector, module)

    assert vegazp == 25.3677
    assert photflam == 4.1652e-21
    assert photfnu == 5.4949e-31
    assert pivot == 1.9887



#! /usr/bin/env python

'''
test to combine crop_mosaic.py and blot_image.py
'''

from . import crop_mosaic, blot_image

crop = crop_mosaic.Extraction()
crop.mosaicfile = 'goodss_candels_udf_wfc3_f105w_010mas_test_drz.fits'
crop.center_ra = 53.122751
crop.center_dec = -27.805089
crop.channel = 'short'
crop.extract()

#crop.cropped now has the datamodel with the mosaic piece...

blot = blot_image.Blot()
blot.blotfile = crop.cropped
blot.detector = ['A1','A1']
blot.center_ra = [53.122751,53.122751]
blot.center_dec = [-27.805089,-27.805089]
blot.pav3 = [0.,12.]
blot.blot()




#!/usr/stsci/pyssg/Python-2.7/bin/python

"""
polynomial.py

Created by Colin Cox on 2013-04-29.
Based on surfit.pro
"""

from math import *

import scipy
from scipy import linalg
import matplotlib.pyplot as P


def triangle(A, order):
    '''Print coefficients in triangular layout'''
    k = 0
    for i in range(order+1):
        for j in range(i+1):
            #print('%12.5e' %A[k], end=' ')
            print('{:f.3}'.format(A[k]))
            k += 1
        print()


def triangulate(A, order):
    '''Convert linear array to 2-D array with triangular coefficient layout'''
    AT = scipy.zeros((order+1, order+1))
    k = 0
    for i in range(order+1):
        for j in range(i+1):
            AT[i, j] = A[k]
            k += 1
    return AT


def flatten(A, order):
    '''Convert triangular layout to linear array'''
    terms = (order+1)*(order+2)//2
    AF = scipy.zeros(terms)
    k = 0
    for i in range(order+1):
        for j in range(i+1):
            AF[k] = A[i, j]
            k += 1
    return AF


def reorder(A, B, verbose=False) :
    '''Reorder Sabatke coefficients to my convention'''
    order = 5
    terms = (order+1)*(order+2)//2
    Aarray = scipy.zeros((order+1, order+1))
    Barray = scipy.zeros((order+1, order+1))
    #Carray = scipy.zeros((order+1, order+1))
    #Darray = scipy.zeros((order+1, order+1))
    k1 = 0
    for i in range(order+1):
        for j in range(order+1-i):
            Aarray[j, i] = A[k1]
            Barray[j, i] = B[k1]
            #Carray[j, i] = C[k1]
            #Darray[j, i] = D[k1]
            k1 += 1

    A2 = scipy.zeros((terms))
    B2 = scipy.zeros((terms))
    #C2 = scipy.zeros((terms))
    #D2 = scipy.zeros((terms))
    k2 = 0
    for i in range(order+1):
        for j in range(i+1):
            A2[k2] = Aarray[j, i-j]
            B2[k2] = Barray[j, i-j]
            #C2[k2] = Carray[j, i-j]
            #D2[k2] = Darray[j, i-j]
            k2 += 1

    if verbose:
        print('A')
        triangle(A2, order)
        print('\nB')
        triangle(B2, order)
        #print '\nC'
        #polyfit(C2, order)
        #print '\nD'
        #triangle(D2, order)

    return (A2, B2) #, C2, D2)


def poly(a, x, y, order):
    pol = 0.0
    k = 0 # index for coefficients
    for i in range(order+1):
        for j in range(i+1):
            pol = pol + a[k]*x**(i-j)*y**j
            k+=1
    return pol


def dpdx(a, x, y, order): # Differential wrt x
    dpdx = 0.0
    k = 1 # index for coefficients
    for i in range(1, order+1):
        for j in range(i+1):
            if i-j > 0: dpdx = dpdx + (i-j)*a[k]*x**(i-j-1)*y**j
            k+=1
    return dpdx


def dpdy(a, x, y, order): # Differential wrt y
    dpdy = 0.0
    k = 1 # index for coefficients
    for i in range(1, order+1):
        for j in range(i+1):
            if j > 0: dpdy = dpdy + j*a[k]*x**(i-j)*y**(j-1)
            k+=1
    return dpdy


def jacob(a, b, x, y, order):
    '''Calculation of Jacobean, or relative area'''
    j = dpdx(a, x, y, order)*dpdy(b, x, y, order) - dpdx(b, x, y, order)*dpdy(a, x, y, order)
    j = scipy.fabs(j)
    return j


def invert(a, b, u, v, n, verbose=False):
    '''Given that order n polynomials of (x, y) have the result (u, v), find (x, y)
    Newton Raphson method in two dimensions'''
    tol = 1.0e-10
    err = 1.0
    #Initial guesses - Linear approximation
    det = a[1]*b[2] - a[2]*b[1]
    x0 = (b[2]*u-a[2]*v)/det
    y0 = (-b[1]*u + a[1]*v)/det
    if verbose: print('Initial guesses', x0, y0)
    x = x0
    y = y0
    X = scipy.array([x, y])
    iter = 0
    while err > tol:
        f1 = scipy.array([poly(a, x, y, n)-u, poly(b, x, y, n)-v])
        j = scipy.array([[dpdx(a, x, y, n), dpdy(a, x, y, n)], [dpdx(b, x, y, n), dpdy(b, x, y, n)]])
        invj = scipy.linalg.inv(j)
        X = X - scipy.dot(invj, f1)
        if verbose: print('[X1, Y1]', X)
        x1 = X[0]
        y1 = X[1]
        err = hypot(x-x1, y-y1)
        if verbose: print('Error %10.2e' % err)
        [x, y] = [x1, y1]
        iter += 1

    return  (x, y, err, iter)


def choose(n, r):
    '''The number of ways of choosing r items from n'''
    if n<0 or r <0:
        print('Negative values not allowed')
        return 0
    if r > n:
        print('r must not be greater than n')
        return 0

    combin = 1
    if r > n/2: r1 = n-r
    else: r1 = r
    for k in range(r1): combin = combin*(n-k)//(k+1)
    return combin


def ShiftCoeffs(a, xshift, yshift, order, verbose=False):
    '''Calculate coefficients of polynomial when shifted to new origin'''

    # First place in triangular layout
    at = scipy.zeros([order+1, order+1])
    atshift = scipy.zeros([order+1, order+1])
    ashift = scipy.zeros([len(a)]) # Copy shape of a
    k = 0
    for p in range(order+1):
        for q in range(p+1):
            at[p, q] = a[k]
            k+=1

    # Apply shift
    for p in range(order+1):
        for q in range(p+1):
            if verbose: print("A'%1d%1d" %(p, q))
            for i in range(p, order+1):
                for j in range(q, i+1-(p-q)):
                    f = choose(j, q)*choose(i-j, p-q)
                    atshift[p, q] = atshift[p, q] + f*xshift**((i-j)-(p-q))*yshift**(j-q)*at[i, j]
                    if verbose: print('%2d A(%1d, %1d) x^%1d y^%1d' %(f, i , j, i-j-(p-q), (j-q)))
            if verbose: print()

    # Put back in linear layout
    k = 0
    for p in range(order+1):
        for q in range(p+1):
            ashift[k] = atshift[p, q]
            k+=1

    return ashift


def polyfit(u, x, y, order):
    '''Fit polynomial to a set of u values on an x, y grid
    u is a function u(x, y) being a polynomial of the form
    u = a[i, j] x**(i-j) y**j. x and y can be on a grid or be arbitrary values'''

    # First set up x and y powers for each coefficient
    px = []
    py = []
    for i in range(order+1):
        for j in range(i+1):
            px.append(i-j)
            py.append(j)
    terms = len(px)
    #print terms, ' terms for order ', order
    #print px
    #print py

    # Make up matrix and vector
    vector = scipy.zeros((terms))
    mat = scipy.zeros((terms, terms))
    for i in range(terms):
        vector[i] = (u*x**px[i]*y**py[i]).sum()
        for j in range(terms):
            mat[i, j] =  (x**px[i]*y**py[i]*x**px[j]*y**py[j]).sum()

    #print 'Vector', vector
    #print 'Matrix'
    #print mat
    imat = linalg.inv(mat)
    #print 'Inverse'
    #print imat
    # Check that inversion worked
    #print scipy.dot(mat, imat)
    coeffs = scipy.dot(imat, vector)
    return coeffs


def polyfit2(u, x, y, order):
    '''Fit polynomial to a set of u values on an x, y grid
    u is a function u(x, y) being a polynomial of the form
    u = a[i, j]x**(i-j)y**j. x and y can be on a grid or be arbitrary values
    This version uses solve instead of matrix inversion'''

    # First set up x and y powers for each coefficient
    px = []
    py = []
    for i in range(order+1):
        for j in range(i+1):
            px.append(i-j)
            py.append(j)
    terms = len(px)
    #print terms, ' terms for order ', order
    #print px
    #print py

    # Make up matrix and vector
    vector = scipy.zeros((terms))
    mat = scipy.zeros((terms, terms))
    for i in range(terms):
        vector[i] = (u*x**px[i]*y**py[i]).sum() # Summing over all x, y
        for j in range(terms):
            mat[i, j] =  (x**px[i]*y**py[i]*x**px[j]*y**py[j]).sum()

    coeffs = linalg.solve(mat, vector)
    return coeffs


def testpoly():
    [x, y] = scipy.mgrid[0:10, 0:10]
    #print 'X'
    #print x
    #print 'Y'
    #print y
    u = scipy.zeros((10, 10))
    v = scipy.zeros((10, 10))
    # Random polynomials
    a0 = scipy.random.rand(1)
    a1 = 0.1*(scipy.random.rand(2)-0.5)
    a2 = 0.01*(scipy.random.rand(3)-0.5)
    a = scipy.concatenate((a0, a1))
    a = scipy.concatenate((a, a2))
    a[2] = 0.01*a[2]
    print('A coefficients')
    print(a)
    b0 = scipy.random.rand(1)
    b1 = 0.1*(scipy.random.rand(2)-0.5)
    b2 = 0.01*(scipy.random.rand(3)-0.5)
    b = scipy.concatenate((b0, b1))
    b = scipy.concatenate((b, b2))
    b[1] = 0.01*b[1]
    print('B coeffcicients')
    print(b)
    for i in range(10):
        for j in range(10):
            u[i, j] = poly(a, x[i, j], y[i, j], 2) #+ scipy.random.normal(0.0, 0.01)
            v[i, j] = poly(b, x[i, j], y[i, j], 2)  #+ scipy.random.normal(0.0, 0.01)
    #print z
    s1 = polyFit2(u, x, y, 2)
    s2 = polyFit2(v, x, y, 2)
    print('S1', s1)
    print('S2', s2)
    uc = poly(s1, x, y, 2)
    vc = poly(s2, x, y, 2)

    P.figure(1)
    P.clf()
    P.grid(True)
    P.plot(u, v, 'gx')
    P.plot(uc, vc, 'r+')


def FlipX(A, order):
    """Change sign of all coefficients with odd x power"""
    terms = (order+1)*(order+2)//2
    AF = scipy.zeros(terms)
    k = 0
    for i in range(order+1):
        for j in range (i+1):
            AF[k] = (-1)**(i-j)*A[k]
            k += 1
    return  AF


def FlipY(A, order):
    """Change sign of all coefficients with odd y power"""
    terms = (order+1)*(order+2)//2
    AF = scipy.zeros(terms)
    k = 0
    for i in range(order+1):
        for j in range (i+1):
            AF[k] = (-1)**(j)*A[k]
            k += 1
    return AF


def FlipXY(A, order):
    "Change sign for coeffs where sum of x and y powers is odd"
    terms = (order+1)*(order+2)//2
    AF = scipy.zeros(terms)
    k = 0
    for i in range(order+1):
        for j in range (i+1):
            AF[k] = (-1)**(i)*A[k]
            k += 1
    return AF


def RotateCoeffs(a, theta, order, verbose=False):
    '''Rotate axes of coefficients by theta degrees'''
    c = cos(radians(theta))
    s = sin(radians(theta))

    # First place in triangular layout
    at = scipy.zeros([order+1, order+1])
    k = 0
    for m in range(order+1):
        for n in range(m+1):
            at[m, n] = a[k]
            k+=1

    # Apply rotation
    atrotate = scipy.zeros([order+1, order+1])
    arotate = scipy.zeros([len(a)]) # Copy shape of a
    for m in range(order+1):
        for n in range(m+1):
            for mu in range(0, m-n+1):
                for j in range(m-n-mu, m-mu+1):
                    factor = (-1)**(m-n-mu)*choose(m-j, mu)*choose(j, m-n-mu)
                    cosSin = c**(j+2*mu-m+n)*s**(2*m-2*mu-j-n)
                    atrotate[m, n] = atrotate[m, n] + factor*cosSin*at[m, j]
                    if verbose: print(m, n, j, factor, 'cos^', j+2*mu-m+n, 'sin^', 2*m-2*mu-j-n, ' A', m, j)
    # Put back in linear layout
    k = 0
    for m in range(order+1):
        for n in range(m+1):
            arotate[k] = atrotate[m, n]
            k+=1

    return arotate


def TransCoeffs(A, a, b, c, d, order, verbose=False):
    '''Transform polynomial coefficients to allow for
    xp = a*x + b*y
    yp = c*x + d*y'''

    A1 = scipy.zeros((order+1, order+1))
    A2 = scipy.zeros((order+1, order+1))
    ncoeffs = (order+1)*(order+2)//2
    if verbose: print(ncoeffs, 'coefficients for order', order)
    AT = scipy.zeros((ncoeffs))

    # First place A in triangular layout
    k = 0
    for i in range(order+1):
        for j in range(i+1):
            A1[i, j] = A[k]
            k+=1

    for m in range(order+1):
        for n in range(m+1):
            if verbose: print('\nM, N', m, n)
            for mu in range(m-n+1):
                for j in range(m-n-mu, m-mu+1):
                    if verbose: print('J, MU', j, mu)
                    if verbose: print('Choose', m-j, mu, 'and', j, m-n-mu)
                    factor = choose(m-j, mu)*choose(j, m-n-mu)
                    A2[m, n] += factor*a**mu*b**(m-j-mu)*c**(m-n-mu)*d**(mu+j-m+n)*A1[m, j]
                    if verbose: print(m, j, ' Factor', factor)
    # Restore A2 to flat layout in AT
    k = 0
    for m in range(order+1):
        for n in range(m+1):
            AT[k] = A2[m, n]
            k += 1
    return AT


def TwoStep(A, B, a, b, order):
    """Change coefficients when
    xp = a[0] + a[1].x + a[2].y
    yp = b[0] + b[1].x + b[2].y"""
    terms = (order+1)*(order+2)//2
    A2 = scipy.zeros((order+1, order+1))
    B2 = scipy.zeros((order+1, order+1))

    k=0
    for i in range(order+1):
        for j in range(i+1):
            for alpha in range(i-j+1):
                for beta in range(i-j-alpha+1):
                    f1 = choose(i-j, alpha)*choose(i-j-alpha, beta)*a[0]**(i-j-alpha-beta)*a[1]**alpha*a[2]**beta
                    for gamma in range(j+1):
                        for delta in range(j-gamma+1):
                            f2 = choose(j, gamma)*choose(j-gamma, delta)*b[0]**(j-gamma-delta)*b[1]**gamma*b[2]**delta
                            A2[alpha+beta+gamma+delta, beta+delta] += A[k]*f1*f2
                            B2[alpha+beta+gamma+delta, beta+delta] += B[k]*f1*f2
            k += 1

    # Flatten A@ and B2
    k = 0
    Aflat = scipy.zeros(terms)
    Bflat = scipy.zeros(terms)
    for i in range(order+1):
        for j in range(i+1):
            Aflat[k] = A2[i, j]
            Bflat[k] = B2[i, j]
            k += 1
    return (Aflat, Bflat)


def TestTwoStep():
    A = scipy.array([10.0, 2.0, 0.1, 0.01, -0.02, 0.03])
    B = scipy.array([4.0, 1.8, 0.2, 0.02, 0.03, -0.02])
    a = scipy.array([1.0, 0.5, 0.1])
    b = scipy.array([2.0, 0.2, 0.6])
    print('\nA')
    triangle(A, 2)
    print('B')
    triangle(B, 2)
    print('a\n', a)
    print('b\n', b)
    (A2, B2) = TwoStep(A, B, a, b, 2)
    print('\nA2')
    triangle(A2, 2)
    print('B2')
    triangle(B2, 2)

    # Now do a test calculation
    (x, y) = (10, 5)
    xp = a[0] + a[1]*x + a[2]*y
    yp = b[0] + b[1]*x + b[2]*y
    print('x, y', x, y)
    print('xp, yp', xp, yp)

    u = poly(A, xp, yp, 2)
    v = poly(B, xp, yp, 2)
    up = poly(A2, x, y, 2)
    vp = poly(B2, x, y, 2)
    print('Two step', u, v)
    print('One step', up, vp)
    return



#! /usr/bin/env python

'''
Module for reading in a given fits file using either RampModel
or astropy, packaging information into a common format,
and returning

metadata needed within the simulator (these need to be in
a common format regardless of read-in method). Other meta
data can stay in whatever format produced when it is read
in, under the assmption that it will be written using the
same format

read pattern - exposure.readpatt READPATT
nints - exposure.nints NINTS
ngroups - exposure.ngroups NGROUPS
nframes - exposure.nframes NFRAMES
nskip - exposure.nskip NSKIP
groupgap - exposure.groupgap GROUPGAP
exptype = exposure.type EXP_TYPE
detector - instrument.detector DETECTOR
instrument - instrument.name INSTRUME
'''
import numpy as np
from astropy.io import fits
from jwst.datamodels import RampModel


class Read_fits():
    def __init__(self):
        self.translate = {}
        self.translate['READPATT'] = 'exposure.readpatt'
        self.translate['NINTS'] = 'exposure.nints'
        self.translate['NGROUPS'] = 'expsoure.ngroups'
        self.translate['NFRAMES'] = 'exposure.nframes'
        self.translate['NSKIP'] = 'exposure.nskip'
        self.translate['GROUPGAP'] = 'exposure.groupgap'
        self.translate['EXP_TYPE'] = 'exposure.type'
        self.translate['DETECTOR'] = 'instrument.detector'
        self.translate['INSTRUME'] = 'instrument.name'
        self.translate['FASTAXIS'] = 'subarray.fastaxis'
        self.translate['SLOWAXIS'] = 'subarray.slowaxis'

    def rampmodel_to_obj(self):
        # convert a RampModel instance to a read_fits object
        self.data = self.model.data
        self.zeroframe = self.model.zeroframe
        self.sbAndRefpix = None
        self.zero_sbAndRefpix = None

        self.header = {}
        for key in self.translate:
            try:
                self.header[key] = self.model.meta[self.translate[key]]
            except:
                self.header[key] = None

    def read_astropy(self):

        h = fits.open(self.file)

        self.data = None
        self.zeroframe = None
        self.sbAndRefpix = None
        self.zero_sbAndRefpix = None
        for i in range(len(h)):
            name = h[i].name
            if name == 'SCI':
                self.data = h[i].data
            if name == 'ZEROFRAME':
                self.zeroframe = h[i].data
            if name == 'SBANDREFPIX':
                self.sbAndRefpix = h[i].data
            if name == 'ZEROSBANDREFPIX':
                self.zero_sbAndRefpix = h[i].data

        #to match what happens with the RampModel version,
        #populate any of the remaining None extensions with
        #arrays of zeros
        #if self.zeroframe is None:
        #    self.zeroframe = np.zeros((ngroup,ny,nx),dtype=np.float)

        self.header = {}
        for key in self.translate:
            try:
                self.header[key] = h[0].header[key]
            except:
                self.header[key] = None

    def read_datamodel(self):

        h = RampModel(self.file)

        #remove any non-pipeline related keywords (e.g. CV3 temps/voltages)
        if 'extra_fits' in dir(h):
            h.__delattr__('extra_fits')

        self.data = h.data

        #Currently a bug in level1bmodel when zeroframe is
        #not present and a cube of zeros is returned
        #If the datamodel returns a default zeroframe of all
        #zeros, then set it to None here
        #self.zeroframe = h.zeroframe
        if np.all(h.zeroframe == 0):
            print("Zeroframe in {}".format(self.file))
            print("All zeros. Returning None.")
            self.zeroframe = None
        else:
            self.zeroframe = h.zeroframe

        self.sbAndRefpix = None

        self.header = {}
        for key in self.translate:
            try:
                self.header[key] = h.meta[self.translate[key]]
            except:
                self.header[key] = None

    def insert_into_datamodel(self,subfile):
        #read in a dummy/substitute file as a datamodel,
        #and insert the data and self.header metadata
        #into it

        h = RampModel(subfile)
        h.data = self.data
        try:
            h.zeroframe = self.zeroframe
        except:
            pass

        h.err = np.zeros_like(self.data)
        h.groupdq = np.zeros_like(self.data)
        nint,ng,ny,nx = self.data.shape
        h.pixeldq = np.zeros((ny,nx))

        h.meta.exposure.readpatt = self.header['READPATT']
        h.meta.exposure.nints = self.header['NINTS']
        h.meta.exposure.ngroups = self.header['NGROUPS']
        h.meta.exposure.nframes = self.header['NFRAMES']
        h.meta.exposure.nskip = self.header['NSKIP']
        h.meta.exposure.groupgap = self.header['GROUPGAP']
        h.meta.exposure.type = self.header['EXP_TYPE']
        h.meta.instrument.detector = self.header['DETECTOR']
        h.meta.instrument.name = self.header['INSTRUME']
        h.meta.subarray.fastaxis = self.header['FASTAXIS']
        h.meta.subarray.slowaxis = self.header['SLOWAXIS']

        return h
